from meloetta.frameworks.porygon.buffer import ReplayBuffer
from meloetta.frameworks.porygon.actor import PorygonActor
from meloetta.frameworks.porygon.model import PorygonModel
from meloetta.frameworks.porygon.learner import PorygonLearner
from meloetta.frameworks.porygon.config import PorygonConfig
from meloetta.frameworks.porygon import utils



from typing import NamedTuple


class _Config(NamedTuple):
    @classmethod
    def from_dict(self, d: dict):
        return self(**d)


class SideEncoderConfig(_Config):
    model_size: int = 128
    num_layers: int = 3
    num_heads: int = 2
    key_size: int = model_size
    value_size: int = model_size // 2
    resblocks_num_before: int = 1
    resblocks_num_after: int = 1
    resblocks_hidden_size: int = key_size
    use_layer_norm: bool = True

    entity_embedding_dim: int = model_size
    output_dim: int = 512


class ScalarEncoderConfig(_Config):
    embedding_dim: int = SideEncoderConfig.output_dim


class WeatherEncoderConfig(_Config):
    embedding_dim: int = SideEncoderConfig.output_dim


class EncoderConfig(_Config):
    side_encoder_config: SideEncoderConfig = SideEncoderConfig()
    scalar_encoder_config: ScalarEncoderConfig = ScalarEncoderConfig()
    weather_encoder_config: WeatherEncoderConfig = WeatherEncoderConfig()


class CoreConfig(_Config):
    side_encoder_dim: int = SideEncoderConfig.output_dim
    scalar_encoder_dim: int = ScalarEncoderConfig.embedding_dim
    weather_encoder_dim: int = WeatherEncoderConfig.embedding_dim

    raw_embedding_dim: int = SideEncoderConfig.output_dim
    hidden_dim: int = raw_embedding_dim
    num_layers: int = 2


class ValueHeadConfig(_Config):
    state_embedding_dim: int = CoreConfig.hidden_dim
    hidden_dim: int = state_embedding_dim
    num_resblocks: int = 2


class ActionTypeHeadConfig(_Config):
    state_embedding_dim: int = CoreConfig.hidden_dim
    num_action_types: int = 3
    residual_dim: int = state_embedding_dim


class FlagHeadConfig(_Config):
    autoregressive_embedding_dim: int = CoreConfig.hidden_dim
    hidden_dim: int = autoregressive_embedding_dim


class MoveHeadConfig(_Config):
    entity_embedding_dim: int = SideEncoderConfig.entity_embedding_dim
    autoregressive_embedding_dim: int = CoreConfig.hidden_dim
    key_dim: int = entity_embedding_dim // 4


class SwitchHeadConfig(_Config):
    entity_embedding_dim: int = SideEncoderConfig.entity_embedding_dim
    autoregressive_embedding_dim: int = CoreConfig.hidden_dim
    key_dim: int = entity_embedding_dim // 4


class TargetHeadConfig(_Config):
    pass


class PolicyHeadsConfig(_Config):
    action_type_head_config: ActionTypeHeadConfig = ActionTypeHeadConfig()
    flag_head_config: FlagHeadConfig = FlagHeadConfig()
    move_head_config: MoveHeadConfig = MoveHeadConfig()
    switch_head_config: SwitchHeadConfig = SwitchHeadConfig()
    target_head_config: TargetHeadConfig = TargetHeadConfig()


def _scale(config: _Config, factor: float):
    datum = {}
    for key in config.__annotations__.keys():
        value = getattr(config, key)
        if hasattr(value, "__annotations__"):
            datum[key] = _scale(value, factor)
        elif isinstance(value, int) or isinstance(value, float):
            datum[key] = max(int(value * factor), 1)

    return config.from_dict(datum)


class NAshKetchumModelConfig(_Config):
    encoder_config: EncoderConfig = EncoderConfig()
    policy_heads_config: PolicyHeadsConfig = PolicyHeadsConfig()
    resnet_core_config: CoreConfig = CoreConfig()
    value_head_config: ValueHeadConfig = ValueHeadConfig()

    @classmethod
    def from_scale(self, scale: float) -> "NAshKetchumModelConfig":
        return _scale(self(), scale)


def main():
    config = NAshKetchumModelConfig.from_scale(0.2)
    print(config.encoder_config.side_encoder_config.entity_embedding_dim)


if __name__ == "__main__":
    main()



import torch

from typing import Callable, Tuple, Union, Dict, List, Any

State = Dict[str, Union[None, torch.Tensor]]
Choices = Dict[str, Dict[str, Tuple[Callable, List[Any], Dict[str, Any]]]]
Battle = Dict[str, Dict[str, Dict[str, Any]]]



import enum
import asyncio
import threading

_global_lock = threading.Lock()

from asyncio import Condition


class BrokenBarrierError(RuntimeError):
    """Barrier is broken by barrier.abort() call."""


class _BarrierState(enum.Enum):
    FILLING = "filling"
    DRAINING = "draining"
    RESETTING = "resetting"
    BROKEN = "broken"


class Barrier:
    """Asyncio equivalent to threading.Barrier
    Implements a Barrier primitive.
    Useful for synchronizing a fixed number of tasks at known synchronization
    points. Tasks block on 'wait()' and are simultaneously awoken once they
    have all made their call.
    """

    def __init__(self, parties):
        """Create a barrier, initialised to 'parties' tasks."""
        if parties < 1:
            raise ValueError("parties must be > 0")

        self._cond = Condition()  # notify all tasks when state changes

        self._parties = parties
        self._state = _BarrierState.FILLING
        self._count = 0  # count tasks in Barrier

    _loop = None

    def _get_loop(self):
        loop = asyncio.get_running_loop()

        if self._loop is None:
            with _global_lock:
                if self._loop is None:
                    self._loop = loop
        # if loop is not self._loop:
        #     raise RuntimeError(f"{self!r} is bound to a different event loop")
        return loop

    def __repr__(self):
        res = super().__repr__()
        extra = f"{self._state.value}"
        if not self.broken:
            extra += f", waiters:{self.n_waiting}/{self.parties}"
        return f"<{res[1:-1]} [{extra}]>"

    async def __aenter__(self):
        # wait for the barrier reaches the parties number
        # when start draining release and return index of waited task
        return await self.wait()

    async def __aexit__(self, *args):
        pass

    async def wait(self):
        """Wait for the barrier.
        When the specified number of tasks have started waiting, they are all
        simultaneously awoken.
        Returns an unique and individual index number from 0 to 'parties-1'.
        """
        async with self._cond:
            await self._block()  # Block while the barrier drains or resets.
            try:
                index = self._count
                self._count += 1
                if index + 1 == self._parties:
                    # We release the barrier
                    await self._release()
                else:
                    await self._wait()
                return index
            finally:
                self._count -= 1
                # Wake up any tasks waiting for barrier to drain.
                self._exit()

    async def _block(self):
        # Block until the barrier is ready for us,
        # or raise an exception if it is broken.
        #
        # It is draining or resetting, wait until done
        # unless a CancelledError occurs
        await self._cond.wait_for(
            lambda: self._state not in (_BarrierState.DRAINING, _BarrierState.RESETTING)
        )

        # see if the barrier is in a broken state
        if self._state is _BarrierState.BROKEN:
            raise BrokenBarrierError("Barrier aborted")

    async def _release(self):
        # Release the tasks waiting in the barrier.

        # Enter draining state.
        # Next waiting tasks will be blocked until the end of draining.
        self._state = _BarrierState.DRAINING
        self._cond.notify_all()

    async def _wait(self):
        # Wait in the barrier until we are released. Raise an exception
        # if the barrier is reset or broken.

        # wait for end of filling
        # unless a CancelledError occurs
        await self._cond.wait_for(lambda: self._state is not _BarrierState.FILLING)

        if self._state in (_BarrierState.BROKEN, _BarrierState.RESETTING):
            raise BrokenBarrierError("Abort or reset of barrier")

    def _exit(self):
        # If we are the last tasks to exit the barrier, signal any tasks
        # waiting for the barrier to drain.
        if self._count == 0:
            if self._state in (_BarrierState.RESETTING, _BarrierState.DRAINING):
                self._state = _BarrierState.FILLING
            self._cond.notify_all()

    async def reset(self):
        """Reset the barrier to the initial state.
        Any tasks currently waiting will get the BrokenBarrier exception
        raised.
        """
        async with self._cond:
            if self._count > 0:
                if self._state is not _BarrierState.RESETTING:
                    # reset the barrier, waking up tasks
                    self._state = _BarrierState.RESETTING
            else:
                self._state = _BarrierState.FILLING
            self._cond.notify_all()

    async def abort(self):
        """Place the barrier into a 'broken' state.
        Useful in case of error.  Any currently waiting tasks and tasks
        attempting to 'wait()' will have BrokenBarrierError raised.
        """
        async with self._cond:
            self._state = _BarrierState.BROKEN
            self._cond.notify_all()

    @property
    def parties(self):
        """Return the number of tasks required to trip the barrier."""
        return self._parties

    @property
    def n_waiting(self):
        """Return the number of tasks currently waiting at the barrier."""
        if self._state is _BarrierState.FILLING:
            return self._count
        return 0

    @property
    def broken(self):
        """Return True if the barrier is in a broken state."""
        return self._state is _BarrierState.BROKEN



import random
import multiprocessing as mp

from meloetta.room import BattleRoom
from meloetta.actors.base import Actor
from meloetta.actors.types import State, Choices


class RandomActor(Actor):
    def __init__(self, queue: mp.Queue = None):
        self.queue = queue

    def choose_action(
        self,
        state: State,
        room: BattleRoom,
        choices: Choices,
    ):
        random_key = random.choice([key for key, value in choices.items() if value])
        _, (func, args, kwargs) = random.choice(list(choices[random_key].items()))
        return func, args, kwargs

    def post_match(self, room: BattleRoom):
        if self.queue is not None:
            datum = room.get_reward()
            reward = datum["reward"]

            if reward > 0:
                self.queue.put(("random", 1))
            elif reward < 0:
                self.queue.put(("random", 0))



import os
import json
import torch
import traceback
import numpy as np

from copy import deepcopy
from abc import ABC
from tqdm import tqdm
from typing import Callable, List, Dict, Any, Set
from collections.abc import MutableMapping

from meloetta.data import GMAX_MOVES
from meloetta.room import BattleRoom


from transformers import AutoTokenizer, AutoModel

# model_id = "princeton-nlp/sup-simcse-roberta-large"
model_id = "dmis-lab/biobert-v1.1"
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
tokenizer = AutoTokenizer.from_pretrained(model_id)
model = AutoModel.from_pretrained(model_id)
model = model.eval()
# model = model.to(device)


def bin_enc(v, n):
    code = f"0{n+2}b"
    data = [int(b) for b in format(int(v), code)[2:]]
    return np.array(data)


@torch.no_grad()
def vectorize_text(text):
    assert text is not None and text
    encoded = tokenizer([text], padding=True, truncation=True, return_tensors="pt")
    vector = model(**encoded).pooler_output.squeeze()
    return vector.numpy()


def one_hot_enc(v, n):
    return np.eye(n)[v]


def sqrt_one_hot_enc(v, n):
    return np.eye(n)[v]


def bow_enc(vs, n):
    m = np.eye(n)
    return np.stack([m[i] for i in vs]).sum(0)


def secondary_enc(chances, effects, num):
    enc = np.zeros(num)
    for chance, effect in zip(chances, effects):
        enc[effect] = chance
    return enc


def single_enc(value):
    return np.array([1 if value is not None else 0])


def z_score(value, mu, sigma):
    return np.array([(value - mu) / sigma])


def flatten(d):
    items = []
    for k, v in d.items():
        if isinstance(v, MutableMapping):
            items.extend([(f"{k}.{t1}", t2) for t1, t2 in flatten(v).items()])
        else:
            items.append((k, v))
    return dict(items)


def get_nested(d, field):
    subfields = field.split(".")
    v = d
    for sf in subfields:
        v = v.get(sf, {})
        if v is None:
            return v
        if not isinstance(v, dict) and not isinstance(v, np.ndarray):
            return v


def get_schema(data: dict):
    schema: Dict[str, set] = {}
    for sample in data.values():
        sample = flatten(sample)
        for key in list(sample):
            if key not in schema:
                schema[key] = set()

    for key in schema:
        for sample in data.values():
            schema[key].add(json.dumps(get_nested(sample, key)))

    schema = dict(sorted(schema.items()))
    return {key: sorted(value) for key, value in schema.items() if len(value) > 1}


class Dex(ABC):
    FEATURES: Dict[str, Callable]
    schema: Dict[str, List[Any]]
    data: Dict[str, List[Any]]

    def stat_statistics(self, stat: str):
        stats = {}
        raw = [p[stat] for p in self.data.values()]
        stats["mu"] = np.mean(raw)
        stats["sigma"] = np.std(raw)
        return stats

    def vectorize(self, feature: str, value: Any):
        func = self.FEATURES.get(feature)
        schema = self.schema[feature]
        if func.__name__ == "one_hot_enc":
            num = len(schema)
            value = schema.index(json.dumps(value))
            return func(value, num)
        elif "bin_enc" in func.__name__:
            value = value or 0
            num = max(list(map(lambda x: int(float(x)) if x != "null" else 0, schema)))
            num = int(np.ceil(np.log2(num)))
            return func(value, num)
        elif "log_enc" in func.__name__:
            return func(value)
        elif func.__name__ == "sqrt_one_hot_enc":
            num = max(list(map(lambda x: int(x), schema)))
            num = int(np.sqrt(num))
            value = int(np.sqrt(value))
            return func(value, num)
        elif func.__name__ == "bow_enc":
            cats = set()
            for p in schema:
                p = json.loads(p)
                cats.update(p)
            cats = list(sorted(cats))
            num = len(cats)
            return func([cats.index(v) for v in value], num)
        elif func.__name__ == "secondary_enc":
            cats = set()
            for p in schema:
                p = json.loads(p)
                if p is not None:
                    for eff in p:
                        eff = json.dumps(
                            {k: v for k, v in eff.items() if k != "chance"}
                        )
                        cats.add(eff)
            cats = list(sorted(cats))
            num = len(cats)
            value = value or []
            chances = [v.pop("chance", 0) / 100 for v in value]
            effects = [cats.index(json.dumps(v)) for v in value]
            return func(chances, effects, num)
        elif func.__name__ == "z_score":
            splits = feature.split(".")
            if len(splits) > 1:
                _, feature = splits
                mu_sigma = getattr(self, f"base_stat_mu_sigma")[feature]
            else:
                mu_sigma = getattr(self, f"{feature}_mu_sigma")
            norm_value = func(value, mu_sigma["mu"], mu_sigma["sigma"])
            return norm_value
        else:
            return func(value)


class Pokedex(Dex):
    FEATURES = {
        "id": one_hot_enc,
        # "baseSpecies": one_hot_enc,
        "formeid": one_hot_enc,
        "types": bow_enc,
        "evos": lambda v: one_hot_enc(int(True if v else False), 2),
        # "abilities.S": one_hot_enc,
        "baseStats.hp": z_score,
        "baseStats.atk": z_score,
        "baseStats.def": z_score,
        "baseStats.spa": z_score,
        "baseStats.spd": z_score,
        "baseStats.spe": z_score,
        "bst": z_score,
        "weightkg": z_score,
        # "heightm": bin_enc,
    }

    def __init__(self, battle: BattleRoom, table: Dict[str, Any], gen: int):
        self.dex = self.get_dex(table, gen)
        data = {o: battle.get_species(o) for o in self.dex}
        self.data = {k: v for k, v in data.items() if v.get("exists", False)}
        self.raw_schema = get_schema(data)
        self.schema = {k: v for k, v in self.raw_schema.items() if k in self.FEATURES}

        self.base_stat_mu_sigma = self.base_stat_statistics()
        self.heightm_mu_sigma = self.stat_statistics("heightm")
        self.weightkg_mu_sigma = self.stat_statistics("weightkg")
        self.bst_mu_sigma = self.stat_statistics("bst")

        self.abilities = set()
        for d in data.values():
            abilities = set(d.get("abilities", {}).values())
            self.abilities.update(abilities)

        for ability, d in (
            table.get(f"gen{gen}", {}).get("overrideAbilityData", {}).items()
        ):
            if "isNonstandard" in d and d["isNonstandard"] is None:
                self.abilities.add(ability)

        self.moves = set()
        for name in self.dex:
            if name in table["learnsets"]:
                self.moves.update(
                    [
                        move
                        for move, gens in table["learnsets"][name].items()
                        if str(gen) in gens
                    ]
                )

        for move, d in table.get(f"gen{gen}", {}).get("overrideMoveData", {}).items():
            if "isNonstandard" in d and d["isNonstandard"] is None:
                self.moves.add(move)

    def base_stat_statistics(self):
        stats = {}
        for stat in {"hp", "atk", "def", "spa", "spd", "spe"}:
            stats[stat] = {}
            try:
                raw = [p["baseStats"][stat] for p in self.data.values()]
                stats[stat]["mu"] = np.mean(raw)
                stats[stat]["sigma"] = np.std(raw)
            except:
                pass
        return stats

    def get_dex(self, table, gen):
        try:
            pokedex = table[f"gen{gen}"]["overrideTier"]
        except:
            pokedex = table["overrideTier"]
        return sorted(
            [p for p, s in pokedex.items() if s != "Illegal" and "CAP" not in s]
        )


class Movedex(Dex):
    FEATURES = {
        "accuracy": lambda x: np.array([0, 1, int(x) / 100])
        if x == True
        else np.array([1, 0, 1]),
        "basePower": z_score,
        "category": one_hot_enc,
        "critRatio": one_hot_enc,
        "flags.allyanim": single_enc,
        "flags.bite": single_enc,
        "flags.bullet": single_enc,
        "flags.bypasssub": single_enc,
        "flags.charge": single_enc,
        "flags.contact": single_enc,
        "flags.dance": single_enc,
        "flags.defrost": single_enc,
        "flags.distance": single_enc,
        "flags.gravity": single_enc,
        "flags.heal": single_enc,
        "flags.mirror": single_enc,
        "flags.nonsky": single_enc,
        "flags.powder": single_enc,
        "flags.protect": single_enc,
        "flags.pulse": single_enc,
        "flags.punch": single_enc,
        "flags.recharge": single_enc,
        "flags.reflectable": single_enc,
        "flags.slicing": single_enc,
        "flags.snatch": single_enc,
        "flags.sound": single_enc,
        "flags.wind": single_enc,
        "hasCrashDamage": one_hot_enc,
        "heal": one_hot_enc,
        "id": one_hot_enc,
        # "maxMove.basePower": bin_enc,
        "multihit": one_hot_enc,
        "noPPBoosts": one_hot_enc,
        "noSketch": one_hot_enc,
        # "num": one_hot_enc,
        "ohko": one_hot_enc,
        "pp": bin_enc,
        "pressureTarget": one_hot_enc,
        "priority": one_hot_enc,
        "recoil": one_hot_enc,
        # "secondaries": secondary_enc,
        "target": one_hot_enc,
        "type": one_hot_enc,
        # "desc": vectorize_text,
        # "zMove.basePower": bin_enc,
        # "zMove.boost.accuracy": one_hot_enc,
        # "zMove.boost.atk": one_hot_enc,
        # "zMove.boost.def": one_hot_enc,
        # "zMove.boost.evasion": one_hot_enc,
        # "zMove.boost.spa": one_hot_enc,
        # "zMove.boost.spd": one_hot_enc,
        # "zMove.boost.spe": one_hot_enc,
        # "zMove.effect": one_hot_enc,
    }

    def __init__(self, battle: BattleRoom, moves: Set[str], gen: int):
        if gen == 8:
            moves.update(GMAX_MOVES)
        data = {move: battle.get_move(move) for move in moves}

        with open("meloetta/js/data/BattleMovedex.json") as f:
            self.raw_json = json.load(f)

        for k in data:
            data[k]["shortDesc"] = self.raw_json[k].get("shortDesc", "")
            data[k]["desc"] = self.raw_json[k].get("desc", "")

        self.data = {k: v for k, v in data.items() if v.get("exists", False)}

        self.raw_schema = get_schema(data)
        self.schema = {
            k: self.raw_schema[k] for k in self.FEATURES if k in self.raw_schema
        }

        self.basePower_mu_sigma = self.stat_statistics("basePower")

        keys_to_pop = []
        if gen < 7:
            for key in self.schema:
                if "zMove" in key:
                    keys_to_pop.append(key)
        if gen < 8:
            for key in self.schema:
                if "maxMove" in key:
                    keys_to_pop.append(key)

        for key in keys_to_pop:
            self.schema.pop(key)


class Itemdex(Dex):
    FEATURES = {
        "id": one_hot_enc,
        "fling.basePower": one_hot_enc,
        "fling.status": one_hot_enc,
        "fling.volatileStatus": one_hot_enc,
        "isPokeball": one_hot_enc,
        "naturalGift.basePower": bin_enc,
        "naturalGift.type": one_hot_enc,
        "onPlate": one_hot_enc,
        # "desc": vectorize_text,
    }

    def __init__(self, battle: BattleRoom, table: Dict[str, Any], gen: int):
        self.dex = self.get_dex(table, gen)
        data = {o: battle.get_item(o) for o in self.dex}

        with open("meloetta/js/data/BattleItems.json") as f:
            self.raw_json = json.load(f)

        for k in data:
            data[k]["shortDesc"] = self.raw_json.get(k, {}).get("shortDesc", "")
            data[k]["desc"] = self.raw_json.get(k, {}).get("desc", "")

        self.data = {k: v for k, v in data.items() if v.get("exists", False)}
        self.raw_schema = get_schema(data)
        self.schema = {k: v for k, v in self.raw_schema.items() if k in self.FEATURES}

    def get_dex(self, table, gen):
        try:
            itemdex = table[f"gen{gen}"]["items"]
        except:
            itemdex = table["items"]
        return sorted([i for i in itemdex if isinstance(i, str)])


class Abilitydex(Dex):
    FEATURES = {
        "id": one_hot_enc,
        # "isPermanent": one_hot_enc,
        # "desc": vectorize_text,
    }

    def __init__(self, battle: BattleRoom, abilities: List[str], gen: int):
        data = {ability: battle.get_ability(ability) for ability in abilities}

        with open("meloetta/js/data/BattleAbilities.json") as f:
            self.raw_json = json.load(f)

        for k, v in data.items():
            data[k]["shortDesc"] = self.raw_json.get(v["id"], {}).get("shortDesc", "")
            data[k]["desc"] = self.raw_json.get(v["id"], {}).get("desc", "")

        self.data = {k: v for k, v in data.items() if v.get("exists", False)}
        self.raw_schema = get_schema(data)
        self.schema = {k: v for k, v in self.raw_schema.items() if k in self.FEATURES}


def main():
    # os.system("npx prettier -w --tab-width 4 pokemon-showdown-client")

    schema = {}
    for gen in range(1, 10):
        battle = BattleRoom()
        battle.set_gen(gen)

        with open("meloetta/js/data/BattleTeambuilderTable.json", "r") as f:
            teambuilder_table = json.load(f)

        save_dir = f"meloetta/pretrained/gen{gen}"
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)

        pokedex = Pokedex(battle, teambuilder_table, gen)
        movedex = Movedex(battle, pokedex.moves, gen)
        abilitydex = Abilitydex(battle, pokedex.abilities, gen)
        itemdex = Itemdex(battle, teambuilder_table, gen)

        schema[f"gen{gen}"] = {}

        for dex in [
            pokedex,
            movedex,
            itemdex,
            abilitydex,
        ]:
            dex: Dex

            dex_name = type(dex).__name__.lower()
            samples = sorted(dex.data.values(), key=lambda x: x["name"])
            progress = tqdm(samples, desc=f"gen{gen}: " + dex_name)
            try:
                for sample in progress:
                    # progress.set_description(sample.get("name", ""))
                    sample["encs"] = {}

                    for feature in dex.schema:
                        value = get_nested(sample, feature)
                        if feature == "desc":
                            if value is None:
                                value = (
                                    value
                                    or dex.raw_json[sample["name"]].get("shortDesc")
                                    or get_nested(sample, "desc")
                                    or dex.raw_json[sample["name"]].get("desc")
                                )
                        sample["encs"][feature] = dex.vectorize(feature, value)

                    feature_vector = np.concatenate(
                        [sample["encs"][feature] for feature in dex.schema]
                    )
                    sample["feature_vector"] = torch.from_numpy(feature_vector)
                data = torch.stack([sample["feature_vector"] for sample in samples])
                data = torch.cat((torch.zeros_like(data[0]).unsqueeze(0), data), dim=0)
                names = [None] + [sample["name"] for sample in samples]
            except:
                traceback.print_exc()
            else:
                schema[f"gen{gen}"][dex_name] = deepcopy(dex.raw_schema)
                for key, values in schema[f"gen{gen}"][dex_name].items():
                    for index, value in enumerate(values):
                        schema[f"gen{gen}"][dex_name][key][index] = json.loads(value)
                save_path = os.path.join(save_dir, dex_name + ".pt")

                print(f"data.shape: {data.shape}")
                torch.save((names, data), save_path)

    with open("meloetta/pretrained/schema.json", "w") as f:
        json.dump(schema, f)


if __name__ == "__main__":
    main()



import numpy as np

from typing import Sequence, Tuple


class EntropySchedule:
    """An increasing list of steps where the regularisation network is updated.
    Example
      EntropySchedule([3, 5, 10], [2, 4, 1])
      =>   [0, 3, 6, 11, 16, 21, 26, 10]
            | 3 x2 |      5 x4     | 10 x1
    """

    def __init__(
        self,
        *,
        sizes: Sequence[int] = (100,),
        repeats: Sequence[int] = (1,),
    ):
        """Constructs a schedule of entropy iterations.
        Args:
          sizes: the list of iteration sizes.
          repeats: the list, parallel to sizes, with the number of times for each
            size from `sizes` to repeat.
        """
        try:
            if len(repeats) != len(sizes):
                raise ValueError("`repeats` must be parallel to `sizes`.")
            if not sizes:
                raise ValueError("`sizes` and `repeats` must not be empty.")
            if any([(repeat <= 0) for repeat in repeats]):
                raise ValueError("All repeat values must be strictly positive")
            if repeats[-1] != 1:
                raise ValueError(
                    "The last value in `repeats` must be equal to 1, "
                    "ince the last iteration size is repeated forever."
                )
        except ValueError as e:
            raise ValueError(
                f"Entropy iteration schedule: repeats ({repeats}) and sizes ({sizes})."
            ) from e

        schedule = [0]
        for size, repeat in zip(sizes, repeats):
            schedule.extend([schedule[-1] + (i + 1) * size for i in range(repeat)])

        self.schedule = np.array(schedule, dtype=np.int32)

    def __call__(self, learner_step: int) -> Tuple[float, bool]:
        """Entropy scheduling parameters for a given `learner_step`.
        Args:
          learner_step: The current learning step.
        Returns:
          alpha: The mixing weight (from [0, 1]) of the previous policy with
            the one before for computing the intrinsic reward.
          update_target_net: A boolean indicator for updating the target network
            with the current network.
        """

        # The complexity below is because at some point we might go past
        # the explicit schedule, and then we'd need to just use the last step
        # in the schedule and apply the logic of
        # ((learner_step - last_step) % last_iteration) == 0)

        # The schedule might look like this:
        # X----X-------X--X--X--X--------X
        # learner_step | might be here ^    |
        # or there     ^                    |
        # or even past the schedule         ^

        # We need to deal with two cases below.
        # Instead of going for the complicated conditional, let's just
        # compute both and then do the A * s + B * (1 - s) with s being a bool
        # selector between A and B.

        # 1. assume learner_step is past the schedule,
        #    ie schedule[-1] <= learner_step.
        last_size = self.schedule[-1] - self.schedule[-2]
        last_start = (
            self.schedule[-1]
            + (learner_step - self.schedule[-1]) // last_size * last_size
        )
        # 2. assume learner_step is within the schedule.
        start = np.amax(self.schedule * (self.schedule <= learner_step))
        finish = np.amin(
            self.schedule * (learner_step < self.schedule),
            initial=self.schedule[-1],
            where=(learner_step < self.schedule),
        )
        size = finish - start

        # Now select between the two.
        beyond = self.schedule[-1] <= learner_step  # Are we past the schedule?
        iteration_start = last_start * beyond + start * (1 - beyond)
        iteration_size = last_size * beyond + size * (1 - beyond)

        update_target_net = np.logical_and(
            learner_step > 0, np.sum(learner_step == iteration_start)
        )
        alpha = np.minimum(
            (2.0 * (learner_step - iteration_start)) / iteration_size, 1.0
        )

        return alpha, update_target_net



import os
import json
import torch
import traceback

from abc import ABC, abstractmethod

from typing import Any, Dict

from meloetta.vector import VectorizedState
from meloetta.room import BattleRoom
from meloetta.actors.types import State, Choices, Battle


Battle = Dict[str, Dict[str, Dict[str, Any]]]


class Actor(ABC):
    def __call__(
        self,
        env_output: State,
        room: BattleRoom,
        choices: Choices,
        *args,
        **kwargs,
    ) -> Any:
        try:
            return self.choose_action(env_output, room, choices)
        except Exception as e:
            battle_tag = room.battle_tag
            state = room.get_state()
            side = state["side"]
            print(f"{battle_tag}: Error Occured")
            trace = traceback.format_exc()
            print(trace)
            datum = {
                "state": state,
                "choices": {
                    k: {sk: sv[1:] for sk, sv in v.items()} for k, v in choices.items()
                },
                "traceback": trace.split("\n"),
                "state_dict": self.model.state_dict(),
                "tensors": {k: v for k, v in env_output.items() if v is not None},
            }

            if not os.path.exists("errors"):
                os.mkdir("errors")

            torch.save(datum, f"errors/{battle_tag}-{side}.pt")

    @abstractmethod
    def choose_action(
        self,
        env_output: State,
        room: BattleRoom,
        choices: Choices,
    ):
        raise NotImplementedError

    def post_match(self, room: BattleRoom):
        pass

    def get_vectorized_state(
        self, room: BattleRoom, battle: Battle
    ) -> Dict[str, torch.Tensor]:
        return VectorizedState.from_battle(room, battle).to_dict()



import torch

from torch import nn
from torch.nn import functional as F

from typing import Tuple

from meloetta.frameworks.porygon.model import config
from meloetta.frameworks.porygon.model.utils import (
    sqrt_one_hot_matrix,
    power_one_hot_matrix,
    TransformerEncoder,
    GLU,
    ToVector,
)
from meloetta.frameworks.porygon.model.interfaces import SideEncoderOutput
from meloetta.embeddings import (
    AbilityEmbedding,
    PokedexEmbedding,
    MoveEmbedding,
    ItemEmbedding,
)

from meloetta.data import (
    GENDERS,
    VOLATILES,
    STATUS,
    TOKENIZED_SCHEMA,
    ITEM_EFFECTS,
    SIDE_CONDITIONS,
    BattleTypeChart,
)


class PrivatePokemonEmbedding(nn.Module):
    def __init__(self, gen: int = 9, output_dim: int = 128) -> None:
        super().__init__()

        self.gen = gen

        pokedex_embedding = PokedexEmbedding(gen=gen)
        self.pokedex_embedding = nn.Sequential(
            pokedex_embedding, nn.Linear(pokedex_embedding.embedding_dim, output_dim)
        )
        ability_embedding = AbilityEmbedding(gen=gen)
        self.ability_embedding = nn.Sequential(
            ability_embedding, nn.Linear(ability_embedding.embedding_dim, output_dim)
        )
        self.item_embedding = ItemEmbedding(gen=gen)
        self.item_effect_onehot = nn.Embedding.from_pretrained(
            torch.eye(len(ITEM_EFFECTS) + 1)[..., 1:]
        )
        self.item_lin = nn.Linear(
            self.item_embedding.embedding_dim + self.item_effect_onehot.embedding_dim,
            output_dim,
        )
        self.pp_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(64))
        self.move_embedding_ae = MoveEmbedding(gen=gen)
        self.move_embedding = nn.Linear(
            self.move_embedding_ae.embedding_dim,  # +self.pp_sqrt_onehot.embedding_dim ,
            output_dim,
        )
        self.last_move_embedding = nn.Linear(
            self.move_embedding_ae.embedding_dim,  # +self.pp_sqrt_onehot.embedding_dim ,
            output_dim,
        )

        self.active_embedding = nn.Embedding(3, output_dim)
        self.fainted_embedding = nn.Embedding(3, output_dim)
        self.gender_embedding = nn.Embedding(len(GENDERS) + 1, output_dim)

        status_dummy = torch.eye(len(STATUS) + 1)[..., 1:]
        self.status_onehot = nn.Embedding.from_pretrained(status_dummy)
        self.sleep_turns_onehot = nn.Embedding.from_pretrained(torch.eye(4)[..., 1:])
        self.toxic_turns_onehot = nn.Embedding.from_pretrained(
            sqrt_one_hot_matrix(16)[..., 1:]
        )
        self.status_lin = nn.Linear(
            self.status_onehot.embedding_dim
            + self.sleep_turns_onehot.embedding_dim
            + self.toxic_turns_onehot.embedding_dim,
            output_dim,
        )

        num_formes = len(TOKENIZED_SCHEMA[f"gen{gen}"]["pokedex"]["forme"]) + 1
        self.forme_embedding = nn.Embedding(num_formes, output_dim)
        self.level_embedding = nn.Embedding(102, output_dim)

        self.hp_sqrt_onehot = nn.Embedding.from_pretrained(
            sqrt_one_hot_matrix(768 if gen != 8 else 1536)[..., 1:]
        )
        self.stat_sqrt_onehot = nn.Embedding.from_pretrained(
            power_one_hot_matrix(512, 1 / 3)[..., 1:]
        )
        self.stat_lin = nn.Linear(
            2 * self.hp_sqrt_onehot.embedding_dim
            + 1
            + 5 * self.stat_sqrt_onehot.embedding_dim,
            output_dim,
        )

        if gen == 9:
            self.commanding_embedding = nn.Embedding(3, output_dim)
            self.reviving_embedding = nn.Embedding(3, output_dim)
            self.tera_embedding = nn.Embedding(2, output_dim)
            self.teratype_embedding = nn.Embedding(
                len(BattleTypeChart) + 1, output_dim, padding_idx=0
            )

        elif gen == 8:
            self.can_gmax_embedding = nn.Embedding.from_pretrained(torch.eye(3))

        self.encoder = nn.Sequential(
            nn.Linear(output_dim, output_dim),
            nn.ReLU(),
            nn.LayerNorm(output_dim),
            nn.Linear(output_dim, output_dim),
            # nn.ReLU(),
            # nn.LayerNorm(output_dim),
            # nn.Linear(output_dim, output_dim),
        )

    def embed_moves(
        self, move_tokens: torch.Tensor, move_used: torch.Tensor
    ) -> torch.Tensor:
        moves_emb = self.move_embedding_ae(move_tokens)
        move_used_emb = self.pp_sqrt_onehot(move_used)
        moves_emb = torch.cat((moves_emb, move_used_emb), dim=-1)
        return self.move_embedding(moves_emb)

    def embed_item(
        self, item_token: torch.Tensor, item_effect_token: torch.Tensor
    ) -> torch.Tensor:
        item_cat = torch.cat(
            (
                self.item_embedding(item_token),
                self.item_effect_onehot(item_effect_token),
            ),
            dim=-1,
        )
        return self.item_lin(item_cat)

    def discretize(self, x: torch.Tensor):
        T, B, _, S, *_ = x.shape
        x = x.view(T, B, 1, S, -1)
        res = x
        x = F.one_hot(x.view(T, B, 1, S, -1, 16).argmax(-1), 16)
        x = x.view(T, B, 1, S, -1)
        x = res + (x - res).detach()
        return x

    def to_dist(self, x: torch.Tensor):
        T, B, _, S, *_ = x.shape
        x = x.view(T, B, 1, S, -1, 16).softmax(-1)
        x = x.view(T, B, 1, S, -1)
        return x

    def forward(
        self, x: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor,]:
        longs = (x + 1).long()

        name = longs[..., 0]
        forme = longs[..., 1]
        slot = longs[..., 2]
        hp = longs[..., 3]
        maxhp = longs[..., 4]
        hp_ratio = x[..., 5]
        stats = longs[..., 6:11]
        fainted = longs[..., 11]
        active = longs[..., 12]
        level = longs[..., 13]
        gender = longs[..., 14]
        ability = longs[..., 15]
        # base_ability = longs[..., 16]
        item = longs[..., 17]
        prev_item = longs[..., 18]
        item_effect = longs[..., 19]
        # prev_item_effect = longs[..., 20]
        status = longs[..., 21]
        sleep_turns = longs[..., 22]
        toxic_turns = longs[..., 23]
        last_move = longs[..., 24]
        moves = longs[..., 25:29]

        if self.gen == 9:
            terastallized = longs[..., 29]
            teratype = longs[..., 30]
            # times_attacked = longs[..., 31]

        name_emb = self.pokedex_embedding(name)
        forme_emb = self.forme_embedding(forme)

        hp_emb = self.hp_sqrt_onehot(hp)
        maxhp_emb = self.hp_sqrt_onehot(maxhp)
        hp_ratio = hp_ratio.unsqueeze(-1)
        stat_onehot = self.stat_sqrt_onehot(stats).flatten(-2)
        stat_emb = self.stat_lin(
            torch.cat((hp_emb, maxhp_emb, hp_ratio, stat_onehot), dim=-1)
        )

        active_emb = self.active_embedding(active)
        fainted_emb = self.fainted_embedding(fainted)
        gender_emb = self.gender_embedding(gender)
        level_emb = self.level_embedding(level)

        ability_emb = self.ability_embedding(ability)
        # base_ability_emb = self.ability_embedding(base_ability)

        item_emb = self.embed_item(item, item_effect)
        # prev_item_emb = self.embed_item(prev_item, prev_item_effect)

        status_onehot = self.status_onehot(status)
        sleep_turns_onehot = self.sleep_turns_onehot(sleep_turns)
        toxic_turns_onehot = self.toxic_turns_onehot(toxic_turns)
        status_emb = self.status_lin(
            torch.cat((status_onehot, sleep_turns_onehot, toxic_turns_onehot), dim=-1)
        )

        moves_emb = self.move_embedding(self.move_embedding_ae(moves))
        moveset_emb = moves_emb.sum(-2) / (moves > 0).sum(-1, keepdim=True).clamp(min=1)

        last_move_emb = self.last_move_embedding(self.move_embedding_ae(last_move))

        pokemon_emb = (
            name_emb
            + forme_emb
            + stat_emb
            + active_emb
            + fainted_emb
            + gender_emb
            + level_emb
            + ability_emb
            # + base_ability_emb
            + item_emb
            # + prev_item_emb
            + status_emb
            + moveset_emb
            + last_move_emb
        )

        if self.gen == 9:
            pokemon_emb = pokemon_emb + self.teratype_embedding(teratype)
            pokemon_emb = pokemon_emb + self.tera_embedding((terastallized > 0).long())

        mask = (name == 0) | (fainted == 2)

        pokemon_emb = self.encoder(pokemon_emb)

        priv, pub1, pub2 = pokemon_emb.chunk(3, 2)
        priv_mask, pub1_mask, pub2_mask = mask.chunk(3, 2)

        # pokemon_emb = torch.cat(
        #     (
        #         self.discretize(priv),
        #         self.to_dist(pub1),
        #         self.to_dist(pub2),
        #     ),
        #     dim=2,
        # )

        return (priv, pub1, pub2), (priv_mask, pub1_mask, pub2_mask), moves_emb


class SideEncoder(nn.Module):
    def __init__(self, gen: int, n_active: int, config: config.SideEncoderConfig):
        super().__init__()

        self.config = config
        self.gen = gen
        self.n_active = n_active

        self.embedding = PrivatePokemonEmbedding(
            gen=gen, output_dim=config.entity_embedding_dim  # , frozen=True
        )

        # self.transformer = TransformerEncoder(
        #     model_size=config.entity_embedding_dim,
        #     key_size=config.entity_embedding_dim,
        #     value_size=config.entity_embedding_dim,
        #     num_heads=config.transformer_num_heads,
        #     num_layers=config.transformer_num_layers,
        #     resblocks_num_before=config.resblocks_num_before,
        #     resblocks_num_after=config.resblocks_num_after,
        # )

        # priv_layer = nn.TransformerEncoderLayer(
        #     d_model=config.entity_embedding_dim,
        #     nhead=config.transformer_num_heads,
        #     dim_feedforward=config.entity_embedding_dim,
        #     dropout=0,
        #     batch_first=True,
        #     norm_first=True,
        # )
        # self.priv_encoder = nn.TransformerEncoder(
        #     priv_layer,
        #     num_layers=config.transformer_num_layers,
        #     enable_nested_tensor=False,
        # )

        # pub_layer = nn.TransformerEncoderLayer(
        #     d_model=config.entity_embedding_dim,
        #     nhead=config.transformer_num_heads,
        #     dim_feedforward=config.entity_embedding_dim,
        #     dropout=0,
        #     batch_first=True,
        #     norm_first=True,
        # )
        # self.pub_encoder = nn.TransformerEncoder(
        #     pub_layer,
        #     num_layers=config.transformer_num_layers,
        #     enable_nested_tensor=False,
        # )

        self.priv_output = ToVector(
            input_dim=config.entity_embedding_dim,
            output_dim=config.output_dim,
        )

        self.pub_output = ToVector(
            input_dim=config.entity_embedding_dim,
            output_dim=config.output_dim,
        )

        self.boosts_onehot = nn.Embedding.from_pretrained(torch.eye(13))
        self.boosts_lin = nn.Linear(8 * 6, config.output_dim)

        self.toxicspikes_onehot = nn.Embedding.from_pretrained(torch.eye(3)[..., 1:])
        self.spikes_onehot = nn.Embedding.from_pretrained(torch.eye(4)[..., 1:])

        side_con_size = (
            2
            + self.toxicspikes_onehot.embedding_dim
            + self.spikes_onehot.embedding_dim
            + (len(SIDE_CONDITIONS) - 4)
        )

        self.hidden_predic = nn.Sequential(
            nn.Linear(config.output_dim, config.output_dim),
            nn.LayerNorm(config.output_dim),
            nn.ReLU(),
            nn.Linear(config.output_dim, config.output_dim),
        )

        self.active_moves = GLU(
            config.entity_embedding_dim,
            config.entity_embedding_dim,
            config.entity_embedding_dim,
        )

    def encode_boosts(self, boosts: torch.Tensor):
        boosts_embed = self.boosts_onehot(boosts + 6)
        boosts_embed = -boosts_embed[..., :6] + boosts_embed[..., 7:]
        return boosts_embed.flatten(2)

    def encode_volatiles(self, volatiles: torch.Tensor):
        return volatiles.float().flatten(2)

    def encode_side_conditions(self, side_conditions: torch.Tensor):
        stealthrock = side_conditions[..., -4].unsqueeze(-1)
        stickyweb = side_conditions[..., -1].unsqueeze(-1)

        toxicspikes_onehot = self.toxicspikes_onehot(side_conditions[..., -3])
        spikes_onehot = self.spikes_onehot(side_conditions[..., -2])

        other_sc = side_conditions[..., :-4]

        side_condition_embed = torch.cat(
            [
                stealthrock,
                stickyweb,
                toxicspikes_onehot,
                spikes_onehot,
                other_sc,
            ],
            dim=-1,
        )
        return side_condition_embed.flatten(2)

    def moves_given_context(self, active: torch.Tensor, moves: torch.Tensor):
        active = active.unsqueeze(-2)  # .repeat_interleave(4, -2)
        moves = moves[..., : self.n_active, :, :]
        return self.active_moves(active, moves)
        # active_and_moves = torch.cat((moves, active), dim=-1)
        # return F.glu(active_and_moves, dim=-1)

    def forward(
        self,
        side: torch.Tensor,
        boosts: torch.Tensor,
        volatiles: torch.Tensor,
        side_conditions: torch.Tensor,
        wisher: torch.Tensor,
    ) -> SideEncoderOutput:

        (
            (priv, pub1, pub2),
            (priv_mask, pub1_mask, pub2_mask),
            moves_emb,
        ) = self.embedding.forward(side)

        T, B, _, S, *_ = priv.shape
        N = 3

        priv_mask = priv_mask.to(torch.bool)
        priv_mask = priv_mask.view(T * B * 1, S)

        pub_mask = torch.cat((pub1_mask, pub2_mask), dim=2).bool()
        pub_mask = pub_mask.view(T * B * 2, S)

        priv_encoder_mask = priv_mask.clone()
        priv_encoder_mask[..., 0] = 0

        pub_encoder_mask = pub_mask.clone()
        pub_encoder_mask[..., 0] = 0

        priv_embeddings = priv.view(T * B, S, -1)
        # priv_embeddings = self.priv_encoder(
        #     priv_embeddings, src_key_padding_mask=priv_encoder_mask
        # )
        priv_embeddings = priv_embeddings * ~priv_mask.unsqueeze(-1)
        priv_embedding = self.priv_output(priv_embeddings, ~priv_mask)
        priv_embedding = priv_embedding.view(T, B, 1, -1)

        pub_embeddings = torch.cat((pub1, pub2), dim=2).view(T * B * 2, S, -1)
        # pub_embeddings = self.priv_encoder(pub_embeddings, src_key_padding_mask=pub_encoder_mask)
        pub_embeddings = pub_embeddings * ~pub_mask.unsqueeze(-1)
        pub_embedding = self.pub_output(pub_embeddings, ~pub_encoder_mask)
        pub_embedding = pub_embedding.view(T, B, 2, -1)

        side_embedding = torch.cat((priv_embedding, pub_embedding), dim=2)

        boosts = self.encode_boosts(boosts)
        volatiles = self.encode_volatiles(volatiles)
        side_conditions = self.encode_side_conditions(side_conditions)

        side_context = torch.cat((boosts, volatiles, side_conditions), dim=-1)

        priv_embeddings = priv_embeddings.view(T, B, S, -1)

        active = priv_embeddings[..., : self.n_active, :]
        moves = self.moves_given_context(active, moves_emb[:, :, 0])
        switches = priv_embeddings[:, :, :6]

        return SideEncoderOutput(
            side_embedding=side_embedding,
            side_context=side_context,
            targ_hidden=None,
            pred_hidden=None,
            moves=moves,
            switches=switches,
        )



import torch
import torch.nn as nn
import torch.nn.functional as F

from typing import Tuple

from meloetta.frameworks.mewzero.model import config
from meloetta.frameworks.mewzero.model.utils import (
    sqrt_one_hot_matrix,
    binary_enc_matrix,
    TransformerEncoder,
    ToVector,
)
from meloetta.frameworks.mewzero.model.encoders import public_spatial
from meloetta.embeddings import (
    AbilityEmbedding,
    PokedexEmbedding,
    MoveEmbedding,
    ItemEmbedding,
)

from meloetta.data import (
    GENDERS,
    STATUS,
    TOKENIZED_SCHEMA,
    BOOSTS,
    SIDE_CONDITIONS,
    VOLATILES,
    ITEM_EFFECTS,
    BattleTypeChart,
)


class PublicEncoder(nn.Module):
    def __init__(self, gen: int, n_active: int, config: config.PublicEncoderConfig):
        super().__init__()

        self.config = config
        self.gen = gen
        self.n_active = n_active

        # scalar stuff
        self.n_onehot = nn.Embedding.from_pretrained(torch.eye(7))
        self.total_onehot = nn.Embedding.from_pretrained(torch.eye(7))
        sc_min_dur_onehot = nn.Embedding.from_pretrained(torch.eye(10))
        self.sc_min_dur_onehot = nn.Sequential(sc_min_dur_onehot, nn.Flatten(-2))
        sc_max_dur_onehot = nn.Embedding.from_pretrained(torch.eye(10))
        self.sc_max_dur_onehot = nn.Sequential(sc_max_dur_onehot, nn.Flatten(-2))
        self.toxic_spikes_onehot = nn.Embedding.from_pretrained(torch.eye(3))
        self.spikes_onehot = nn.Embedding.from_pretrained(torch.eye(4))
        self.stealthrock_onehot = nn.Embedding.from_pretrained(torch.eye(2))
        self.stickyweb_onehot = nn.Embedding.from_pretrained(torch.eye(2))

        n_special_sc = 4
        normal_sc = len(SIDE_CONDITIONS) - n_special_sc
        sc_lin_in = (
            self.n_onehot.embedding_dim
            + self.n_onehot.embedding_dim
            + self.total_onehot.embedding_dim
            + self.n_onehot.embedding_dim
            + normal_sc
            + (normal_sc * sc_min_dur_onehot.embedding_dim)
            + (normal_sc * sc_min_dur_onehot.embedding_dim)
            + self.toxic_spikes_onehot.embedding_dim
            + self.spikes_onehot.embedding_dim
            + self.stealthrock_onehot.embedding_dim
            + self.stickyweb_onehot.embedding_dim
        )
        self.scalar_lin = nn.Linear(sc_lin_in, config.scalar_embedding_dim)

        self.active_onehot = nn.Embedding.from_pretrained(torch.eye(2))
        self.sideid_onehot = nn.Embedding.from_pretrained(torch.eye(3))
        self.slot_onehot = nn.Embedding.from_pretrained(torch.eye(n_active + 1))
        self.fainted_onehot = nn.Embedding.from_pretrained(torch.eye(3))
        self.gender_onehot = nn.Embedding.from_pretrained(torch.eye(len(GENDERS) + 1))
        self.status_onehot = nn.Embedding.from_pretrained(torch.eye(len(STATUS) + 1))
        self.item_effect_onehot = nn.Embedding.from_pretrained(
            torch.eye(len(ITEM_EFFECTS) + 1)
        )
        self.times_attacked_onehot = nn.Embedding.from_pretrained(torch.eye(8))

        self.forme_onehot = nn.Embedding.from_pretrained(
            torch.eye(len(TOKENIZED_SCHEMA[f"gen{gen}"]["pokedex"]["forme"]) + 1)
        )

        # binaries
        self.level_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(102))
        self.hp_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(102))
        self.toxic_bin = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(17))
        self.sleep_bin = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(5))

        status_embedding_dim = (
            self.status_onehot.embedding_dim
            + self.toxic_bin.embedding_dim
            + self.sleep_bin.embedding_dim
            + 2
        )

        # precomputed embeddings
        self.ability_embedding = AbilityEmbedding(gen=gen)
        self.pokedex_embedding = PokedexEmbedding(gen=gen)
        self.item_embedding = ItemEmbedding(gen=gen)
        item_embedding_dim = (
            2 * self.item_embedding.embedding_dim
            + 2 * self.item_effect_onehot.embedding_dim
        )

        self.move_embedding = MoveEmbedding(gen=gen)

        self.pp_sqrt_onehot = nn.Embedding.from_pretrained(sqrt_one_hot_matrix(66))
        self.move_lin = nn.Linear(
            self.move_embedding.embedding_dim + self.pp_sqrt_onehot.embedding_dim,
            config.entity_embedding_dim,
        )

        hp_embedding_dim = self.hp_sqrt_onehot.embedding_dim + 1
        level_embedding_dim = self.level_sqrt_onehot.embedding_dim + 1

        base_embedding_size = (
            self.pokedex_embedding.embedding_dim
            + self.active_onehot.embedding_dim
            + self.sideid_onehot.embedding_dim
            + self.forme_onehot.embedding_dim
            + self.slot_onehot.embedding_dim
            + hp_embedding_dim
            + self.fainted_onehot.embedding_dim
            + level_embedding_dim
            + self.gender_onehot.embedding_dim
            + 2 * self.ability_embedding.embedding_dim
            + item_embedding_dim
            + status_embedding_dim
            + self.move_embedding.embedding_dim
            + self.times_attacked_onehot.embedding_dim
        )
        if self.gen == 9:
            self.teratype_onehot = nn.Embedding.from_pretrained(
                torch.eye(len(BattleTypeChart) + 1)
            )
            base_embedding_size += self.teratype_onehot.embedding_dim

        self.entity_lin = nn.Linear(base_embedding_size, config.entity_embedding_dim)

        self.transformer = TransformerEncoder(
            key_size=config.entity_embedding_dim,
            value_size=config.entity_embedding_dim,
            num_heads=config.transformer_num_heads,
            num_layers=config.transformer_num_layers,
            resblocks_num_before=config.resblocks_num_before,
            resblocks_num_after=config.resblocks_num_after,
        )
        self.output = ToVector(
            input_dim=config.entity_embedding_dim,
            output_dim=2 * config.entity_embedding_dim,
        )
        self.spatial = public_spatial.PublicSpatialEncoder(
            gen=gen, n_active=n_active, config=config
        )

    def embed_hp(self, hp_ratio: torch.Tensor) -> torch.Tensor:
        hp_ratio = hp_ratio * (hp_ratio >= 0)
        hp_cat = hp_ratio.clamp(min=0) * 100
        hp_emb = self.hp_sqrt_onehot(hp_cat)
        hp_emb = torch.cat([hp_ratio.unsqueeze(-1), hp_emb], dim=-1)
        return hp_emb

    def embed_level(self, level: torch.Tensor) -> torch.Tensor:
        level_bin = self.level_sqrt_onehot(level)
        level_ratio = (level / 100).unsqueeze(-1)
        level_emb = torch.cat((level_bin, level_ratio), dim=-1)
        return level_emb

    def embed_pp(self, pp: torch.Tensor) -> torch.Tensor:
        return self.pp_sqrt_onehot(pp)

    def embed_ability(self, current_and_base_ability: torch.Tensor) -> torch.Tensor:
        return torch.flatten(self.ability_embedding(current_and_base_ability), -2)

    def embed_item(
        self,
        curr_item: torch.Tensor,
        curr_item_effect: torch.Tensor,
        prev_item: torch.Tensor,
        prev_item_effect: torch.Tensor,
    ) -> torch.Tensor:
        item_emb = torch.cat(
            (
                self.item_embedding(curr_item),
                self.item_embedding(prev_item),
                self.item_effect_onehot(curr_item_effect),
                self.item_effect_onehot(prev_item_effect),
            ),
            dim=-1,
        )
        return item_emb

    def embed_status(
        self,
        status: torch.Tensor,
        status_stage: torch.Tensor,
        sleep_turns: torch.Tensor,
        toxic_turns: torch.Tensor,
    ) -> torch.Tensor:
        status_emb = torch.cat(
            (
                self.status_onehot(status),
                (sleep_turns / 5).unsqueeze(-1),
                self.sleep_bin(sleep_turns),
                (toxic_turns / 17).unsqueeze(-1),
                self.toxic_bin(toxic_turns),
            ),
            dim=-1,
        )
        return status_emb

    def embed_last_move(self, last_move: torch.Tensor) -> torch.Tensor:
        return self.move_embedding(last_move)

    def embed_moves(
        self, moves_id: torch.Tensor, moves_pp: torch.Tensor
    ) -> torch.Tensor:
        moves_id_emb = self.move_embedding(moves_id)
        moves_pp_emb = self.embed_pp(moves_pp)
        moves_emb = torch.cat((moves_id_emb, moves_pp_emb), dim=-1)
        return self.move_lin(moves_emb)

    def embed_scalars(
        self,
        n: torch.Tensor,
        total_pokemon: torch.Tensor,
        faint_counter: torch.Tensor,
        side_conditions: torch.Tensor,
        wisher: torch.Tensor,
        stealthrock: torch.Tensor,
        spikes: torch.Tensor,
        toxicspikes: torch.Tensor,
        stickyweb: torch.Tensor,
    ) -> torch.Tensor:
        side_conditions_x = side_conditions + 1
        sc_levels = side_conditions[..., 0]
        sc_min_dur = side_conditions_x[..., 1]
        sc_max_dur = side_conditions_x[..., 2]

        scalars = torch.cat(
            [
                self.n_onehot(n),
                self.total_onehot(total_pokemon),
                self.n_onehot(faint_counter),
                self.n_onehot(wisher + 1),
                sc_levels,
                self.sc_min_dur_onehot(sc_min_dur),
                self.sc_max_dur_onehot(sc_max_dur),
                self.toxic_spikes_onehot(toxicspikes),
                self.spikes_onehot(spikes),
                self.stealthrock_onehot(stealthrock),
                self.stickyweb_onehot(stickyweb),
            ],
            dim=-1,
        )
        return self.scalar_lin(scalars)

    def embed_active(self, active: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        active = active.long()
        active_x = active + 1

        species = active_x[..., 0]
        forme = active_x[..., 1]
        slot = active_x[..., 2]
        hp = active[..., 3]
        fainted = active_x[..., 4]
        level = active_x[..., 5]
        gender = active_x[..., 6]
        current_and_base_ability = active_x[..., 7:9]
        item = active_x[..., 9]
        item_effect = active_x[..., 10]
        prev_item = active_x[..., 11]
        prev_item_effect = active_x[..., 12]
        terastallized = active_x[..., 13]
        status = active_x[..., 14]
        status_stage = active_x[..., 15]
        last_move = active_x[..., 16]
        times_attacked = active_x[..., 17]
        sleep_turns = active_x[..., 18]
        toxic_turns = active_x[..., 19]
        boosts = active[..., 20 : 20 + len(BOOSTS)] + 6
        volatiles = (
            active[..., 20 + len(BOOSTS) : 20 + len(BOOSTS) + len(VOLATILES)]
            .clamp(min=0)
            .float()
        )
        sideid = active_x[..., -17]

        moves = active_x[..., -16:]
        moves = moves.view(*moves.shape[:-1], 8, 2)
        moves_id = moves[..., 0]
        moves_pp = moves[..., 1]
        moves_emb = self.embed_moves(moves_id, moves_pp)

        entity_embed = [
            self.pokedex_embedding(species),
            self.active_onehot(torch.zeros_like(slot)),
            self.sideid_onehot(sideid),
            self.forme_onehot(forme),
            self.slot_onehot(slot),
            self.embed_hp(hp),
            self.fainted_onehot(fainted),
            self.embed_level(level),
            self.gender_onehot(gender),
            self.embed_ability(current_and_base_ability),
            self.embed_item(
                item,
                item_effect,
                prev_item,
                prev_item_effect,
            ),
            self.embed_status(
                status,
                status_stage,
                sleep_turns,
                toxic_turns,
            ),
            self.embed_last_move(last_move),
            self.times_attacked_onehot(times_attacked),
        ]
        if self.gen == 9:
            entity_embed += [self.teratype_onehot(terastallized)]

        entity_embed = torch.cat(entity_embed, dim=-1)
        entity_embed = self.entity_lin(entity_embed)
        entity_embed = entity_embed + moves_emb.sum(-2)

        mask = (species == 0) | (fainted == 2)

        return entity_embed, mask, boosts, volatiles

    def embed_reserve(self, reserve: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        reserve = reserve.long()
        reserve_x = reserve + 1

        species = reserve_x[..., 0]
        forme = reserve_x[..., 1]
        slot = reserve_x[..., 2]
        hp = reserve[..., 3]
        fainted = reserve_x[..., 4]
        level = reserve_x[..., 5]
        gender = reserve_x[..., 6]
        current_and_base_ability = reserve_x[..., 7:9]
        curr_item = reserve_x[..., 9]
        curr_item_effect = reserve_x[..., 10]
        prev_item = reserve_x[..., 11]
        prev_item_effect = reserve_x[..., 12]
        terastallized = reserve_x[..., 13]
        status = reserve_x[..., 14]
        status_stage = reserve_x[..., 15]
        last_move = reserve_x[..., 16]
        times_attacked = reserve_x[..., 17]
        sleep_turns = reserve_x[..., 18]
        toxic_turns = reserve_x[..., 18]
        sideid = reserve_x[..., 20]

        moves = reserve_x[..., 21:]
        moves = moves.view(*moves.shape[:-1], 8, 2)
        moves_id = moves[..., 0]
        moves_pp = moves[..., 1]
        moves_emb = self.embed_moves(moves_id, moves_pp)

        entity_embed = [
            self.pokedex_embedding(species),
            self.active_onehot(torch.ones_like(slot)),
            self.sideid_onehot(sideid),
            self.forme_onehot(forme),
            self.slot_onehot(slot),
            self.embed_hp(hp),
            self.fainted_onehot(fainted),
            self.embed_level(level),
            self.gender_onehot(gender),
            self.embed_ability(current_and_base_ability),
            self.embed_item(
                curr_item,
                curr_item_effect,
                prev_item,
                prev_item_effect,
            ),
            self.embed_status(
                status,
                status_stage,
                sleep_turns,
                toxic_turns,
            ),
            self.embed_last_move(last_move),
            self.times_attacked_onehot(times_attacked),
        ]
        if self.gen == 9:
            entity_embed += [self.teratype_onehot(terastallized)]

        entity_embed = torch.cat(entity_embed, dim=-1)
        entity_embed = self.entity_lin(entity_embed)
        entity_embed = entity_embed + moves_emb.sum(-2)

        mask = (species == 0) | (fainted == 2)

        return entity_embed, mask

    def forward(
        self,
        n: torch.Tensor,
        total_pokemon: torch.Tensor,
        faint_counter: torch.Tensor,
        side_conditions: torch.Tensor,
        wisher: torch.Tensor,
        active: torch.Tensor,
        reserve: torch.Tensor,
        stealthrock: torch.Tensor,
        spikes: torch.Tensor,
        toxicspikes: torch.Tensor,
        stickyweb: torch.Tensor,
    ):
        scalar_emb = self.embed_scalars(
            n,
            total_pokemon,
            faint_counter,
            side_conditions,
            wisher,
            stealthrock,
            spikes,
            toxicspikes,
            stickyweb,
        )

        active_entity_embed, active_mask, boosts, volatiles = self.embed_active(active)
        reserve_entity_embed, reserve_mask = self.embed_reserve(reserve)

        entity_embeddings = torch.cat(
            [
                active_entity_embed,
                reserve_entity_embed,
            ],
            dim=-2,
        )

        T, B, S, L, *_ = entity_embeddings.shape

        mask = torch.cat(
            [
                active_mask,
                reserve_mask,
            ],
            dim=-1,
        )
        mask = mask.bool()
        mask = mask.view(T * B, S * L)

        encoder_mask = mask.clone()
        encoder_mask[..., 0] = False

        entity_embeddings: torch.Tensor = entity_embeddings.view(T * B, S * L, -1)
        entity_embeddings = self.transformer(entity_embeddings, encoder_mask)

        mask = ~mask.unsqueeze(-1)
        entity_embeddings = entity_embeddings * mask
        entity_embedding = self.output(entity_embeddings)

        entity_embeddings = entity_embeddings.view(T * B, S, L, -1)
        mask = mask.view(T * B, S, L, -1)

        spatial_embedding = self.spatial(
            entity_embeddings, mask, boosts, volatiles, scalar_emb
        )
        spatial_embedding = spatial_embedding.view(T, B, -1)

        entity_embedding = entity_embedding.view(T, B, -1)

        return entity_embedding, spatial_embedding


class PublicEncoderV2(nn.Module):
    def __init__(self, gen: int, n_active: int, config: config.PublicEncoderConfig):
        super().__init__()

        self.config = config
        self.gen = gen
        self.n_active = n_active

        # scalar stuff
        self.n_onehot = nn.Embedding.from_pretrained(torch.eye(7))
        self.total_onehot = nn.Embedding.from_pretrained(torch.eye(7))

        sc_min_dur_onehot = nn.Embedding.from_pretrained(torch.eye(10))
        self.sc_min_dur_onehot = nn.Sequential(sc_min_dur_onehot, nn.Flatten(-2))

        sc_max_dur_onehot = nn.Embedding.from_pretrained(torch.eye(10))
        self.sc_max_dur_onehot = nn.Sequential(sc_max_dur_onehot, nn.Flatten(-2))

        self.toxic_spikes_onehot = nn.Embedding.from_pretrained(torch.eye(3))
        self.spikes_onehot = nn.Embedding.from_pretrained(torch.eye(4))
        self.stealthrock_onehot = nn.Embedding.from_pretrained(torch.eye(2))
        self.stickyweb_onehot = nn.Embedding.from_pretrained(torch.eye(2))

        n_special_sc = 4
        normal_sc = len(SIDE_CONDITIONS) - n_special_sc
        sc_lin_in = (
            self.n_onehot.embedding_dim
            + self.n_onehot.embedding_dim
            + self.total_onehot.embedding_dim
            + self.n_onehot.embedding_dim
            + normal_sc
            + (normal_sc * sc_min_dur_onehot.embedding_dim)
            + (normal_sc * sc_min_dur_onehot.embedding_dim)
            + self.toxic_spikes_onehot.embedding_dim
            + self.spikes_onehot.embedding_dim
            + self.stealthrock_onehot.embedding_dim
            + self.stickyweb_onehot.embedding_dim
        )
        self.scalar_lin = nn.Linear(sc_lin_in, config.scalar_embedding_dim)

        self.active_embedding = nn.Embedding(2, config.entity_embedding_dim)

        self.sideid_embedding = nn.Embedding(3, config.entity_embedding_dim)
        self.slot_embedding = nn.Embedding(n_active + 1, config.entity_embedding_dim)
        self.fainted_embedding = nn.Embedding(3, config.entity_embedding_dim)
        self.gender_embedding = nn.Embedding(
            len(GENDERS) + 1, config.entity_embedding_dim
        )

        self.forme_embedding = nn.Embedding(
            len(TOKENIZED_SCHEMA[f"gen{gen}"]["pokedex"]["forme"]) + 1,
            config.entity_embedding_dim,
        )

        # binaries
        self.level_embedding = nn.Embedding(102, config.entity_embedding_dim)

        self.hp_onehot = nn.Embedding.from_pretrained(torch.eye(33))
        self.hp_embedding = nn.Linear(
            self.hp_onehot.embedding_dim + 1, config.entity_embedding_dim, bias=False
        )

        self.status_onehot = nn.Embedding.from_pretrained(torch.eye(len(STATUS) + 1))
        self.toxic_onehot = nn.Embedding.from_pretrained(torch.eye(17))
        self.sleep_onehot = nn.Embedding.from_pretrained(torch.eye(5))
        self.status_embedding = nn.Linear(
            self.status_onehot.embedding_dim
            + self.toxic_onehot.embedding_dim
            + self.sleep_onehot.embedding_dim,
            config.entity_embedding_dim,
            bias=False,
        )

        # ability embeddings
        self.ability_embedding = AbilityEmbedding(gen=gen)
        self.current_ability_lin = nn.Linear(
            self.ability_embedding.embedding_dim,
            config.entity_embedding_dim,
            bias=False,
        )
        self.base_ability_lin = nn.Linear(
            self.ability_embedding.embedding_dim,
            config.entity_embedding_dim,
            bias=False,
        )

        # pokedex embeddings
        pokedex_embedding = PokedexEmbedding(gen=gen)
        self.pokedex_embedding = nn.Sequential(
            pokedex_embedding,
            nn.Linear(
                pokedex_embedding.embedding_dim, config.entity_embedding_dim, bias=False
            ),
        )

        # item embeddings
        self.item_embedding = ItemEmbedding(gen=gen)
        self.item_effect_onehot = nn.Embedding.from_pretrained(
            torch.eye(len(ITEM_EFFECTS) + 1)
        )
        self.curr_item_lin = nn.Linear(
            self.item_effect_onehot.embedding_dim + self.item_embedding.embedding_dim,
            config.entity_embedding_dim,
            bias=False,
        )
        self.prev_item_lin = nn.Linear(
            self.item_effect_onehot.embedding_dim + self.item_embedding.embedding_dim,
            config.entity_embedding_dim,
            bias=False,
        )

        self.move_embedding = MoveEmbedding(gen=gen)

        self.pp_bin = nn.Embedding.from_pretrained(binary_enc_matrix(66))
        self.moveset_lin = nn.Linear(
            self.move_embedding.embedding_dim + self.pp_bin.embedding_dim,
            config.entity_embedding_dim,
        )
        self.prev_move_lin = nn.Linear(
            self.move_embedding.embedding_dim,
            config.entity_embedding_dim,
        )

        if self.gen == 9:
            self.times_attacked_embedding = nn.Embedding(8, config.entity_embedding_dim)
            self.teratype_embedding = nn.Embedding(
                len(BattleTypeChart) + 1, config.entity_embedding_dim
            )

        self.transformer = TransformerEncoder(
            key_size=config.entity_embedding_dim,
            value_size=config.entity_embedding_dim,
            num_heads=config.transformer_num_heads,
            num_layers=config.transformer_num_layers,
            resblocks_num_before=config.resblocks_num_before,
            resblocks_num_after=config.resblocks_num_after,
        )
        self.output = ToVector(
            input_dim=config.entity_embedding_dim,
            output_dim=2 * config.entity_embedding_dim,
        )
        self.spatial = public_spatial.PublicSpatialEncoder(
            gen=gen, n_active=n_active, config=config
        )

    def embed_hp(self, hp_ratio: torch.Tensor) -> torch.Tensor:
        hp_ratio = hp_ratio * (hp_ratio >= 0)
        hp_ratio = hp_ratio.clamp(min=0, max=1) * 100
        hp_emb = self.hp_onehot((hp_ratio * 32 / 100).ceil().long())
        hp_cat = torch.cat([hp_ratio.unsqueeze(-1), hp_emb], dim=-1)
        return self.hp_embedding(hp_cat)

    def embed_level(self, level: torch.Tensor) -> torch.Tensor:
        return self.level_embedding(level)

    def embed_pp(self, pp: torch.Tensor) -> torch.Tensor:
        return self.pp_bin(pp)

    def embed_ability(self, current_and_base_ability: torch.Tensor) -> torch.Tensor:
        current_ability_embedding, base_ability_embedding = torch.chunk(
            self.ability_embedding(current_and_base_ability), 2, -2
        )
        current_ability_embedding = self.current_ability_lin(
            current_ability_embedding.squeeze(-2)
        )
        base_ability_embedding = self.base_ability_lin(
            base_ability_embedding.squeeze(-2)
        )
        return current_ability_embedding + base_ability_embedding

    def embed_item(
        self,
        curr_item: torch.Tensor,
        curr_item_effect: torch.Tensor,
        prev_item: torch.Tensor,
        prev_item_effect: torch.Tensor,
    ) -> torch.Tensor:
        curr_item_cat = torch.cat(
            (self.item_embedding(curr_item), self.item_effect_onehot(curr_item_effect)),
            dim=-1,
        )
        curr_item_embedding = self.curr_item_lin(curr_item_cat)
        prev_item_cat = torch.cat(
            (self.item_embedding(prev_item), self.item_effect_onehot(prev_item_effect)),
            dim=-1,
        )
        prev_item_embedding = self.prev_item_lin(prev_item_cat)
        return curr_item_embedding + prev_item_embedding

    def embed_status(
        self,
        status: torch.Tensor,
        status_stage: torch.Tensor,
        sleep_turns: torch.Tensor,
        toxic_turns: torch.Tensor,
    ) -> torch.Tensor:
        status_concat_onehot = torch.cat(
            (
                self.status_onehot(status),
                self.sleep_onehot(sleep_turns),
                self.toxic_onehot(toxic_turns),
            ),
            dim=-1,
        )
        return self.status_embedding(status_concat_onehot)

    def embed_last_move(self, last_move: torch.Tensor) -> torch.Tensor:
        last_move_onehot = self.move_embedding(last_move)
        return self.prev_move_lin(last_move_onehot)

    def embed_moveset(
        self, moves_id: torch.Tensor, moves_pp: torch.Tensor
    ) -> torch.Tensor:
        moves_id_emb = self.move_embedding(moves_id)
        moves_pp_emb = self.embed_pp(moves_pp)
        moves_emb = torch.cat((moves_id_emb, moves_pp_emb), dim=-1)
        return self.moveset_lin(moves_emb).sum(-2)

    def embed_scalars(
        self,
        n: torch.Tensor,
        total_pokemon: torch.Tensor,
        faint_counter: torch.Tensor,
        side_conditions: torch.Tensor,
        wisher: torch.Tensor,
        stealthrock: torch.Tensor,
        spikes: torch.Tensor,
        toxicspikes: torch.Tensor,
        stickyweb: torch.Tensor,
    ) -> torch.Tensor:
        side_conditions_x = side_conditions + 1
        sc_levels = side_conditions[..., 0]
        sc_min_dur = side_conditions_x[..., 1]
        sc_max_dur = side_conditions_x[..., 2]

        scalars = torch.cat(
            [
                self.n_onehot(n),
                self.total_onehot(total_pokemon),
                self.n_onehot(faint_counter),
                self.n_onehot(wisher + 1),
                sc_levels,
                self.sc_min_dur_onehot(sc_min_dur),
                self.sc_max_dur_onehot(sc_max_dur),
                self.toxic_spikes_onehot(toxicspikes),
                self.spikes_onehot(spikes),
                self.stealthrock_onehot(stealthrock),
                self.stickyweb_onehot(stickyweb),
            ],
            dim=-1,
        )
        return self.scalar_lin(scalars)

    def embed_active(self, active: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        active = active.long()
        active_x = active + 1

        species = active_x[..., 0]
        forme = active_x[..., 1]
        slot = active_x[..., 2]
        hp = active[..., 3]
        fainted = active_x[..., 4]
        level = active_x[..., 5]
        gender = active_x[..., 6]
        current_and_base_ability = active_x[..., 7:9]
        item = active_x[..., 9]
        item_effect = active_x[..., 10]
        prev_item = active_x[..., 11]
        prev_item_effect = active_x[..., 12]
        terastallized = active_x[..., 13]
        status = active_x[..., 14]
        status_stage = active_x[..., 15]
        last_move = active_x[..., 16]
        times_attacked = active_x[..., 17]
        sleep_turns = active_x[..., 18]
        toxic_turns = active_x[..., 19]
        boosts = active[..., 20 : 20 + len(BOOSTS)] + 6
        volatiles = (
            active[..., 20 + len(BOOSTS) : 20 + len(BOOSTS) + len(VOLATILES)]
            .clamp(min=0)
            .float()
        )
        sideid = active_x[..., -17]

        moves = active_x[..., -16:]
        moves = moves.view(*moves.shape[:-1], 8, 2)
        moves_id = moves[..., 0]
        moves_pp = moves[..., 1]

        entity_embed = (
            self.pokedex_embedding(species)
            + self.active_embedding(torch.zeros_like(slot))
            + self.sideid_embedding(sideid)
            + self.forme_embedding(forme)
            + self.slot_embedding(slot)
            + self.fainted_embedding(fainted)
            + self.gender_embedding(gender)
            + self.embed_hp(hp)
            + self.embed_level(level)
            + self.embed_ability(current_and_base_ability)
            + self.embed_item(item, item_effect, prev_item, prev_item_effect)
            + self.embed_status(status, status_stage, sleep_turns, toxic_turns)
            + self.embed_last_move(last_move)
            + self.embed_moveset(moves_id, moves_pp)
        )
        if self.gen == 9:
            entity_embed += self.teratype_embedding(terastallized)
            entity_embed += self.times_attacked_embedding(times_attacked)

        mask = (species == 0) | (fainted == 2)

        return entity_embed, mask, boosts, volatiles

    def embed_reserve(self, reserve: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        reserve = reserve.long()
        reserve_x = reserve + 1

        species = reserve_x[..., 0]
        forme = reserve_x[..., 1]
        slot = reserve_x[..., 2]
        hp = reserve[..., 3]
        fainted = reserve_x[..., 4]
        level = reserve_x[..., 5]
        gender = reserve_x[..., 6]
        current_and_base_ability = reserve_x[..., 7:9]
        curr_item = reserve_x[..., 9]
        curr_item_effect = reserve_x[..., 10]
        prev_item = reserve_x[..., 11]
        prev_item_effect = reserve_x[..., 12]
        terastallized = reserve_x[..., 13]
        status = reserve_x[..., 14]
        status_stage = reserve_x[..., 15]
        last_move = reserve_x[..., 16]
        times_attacked = reserve_x[..., 17]
        sleep_turns = reserve_x[..., 18]
        toxic_turns = reserve_x[..., 18]
        sideid = reserve_x[..., 20]

        moves = reserve_x[..., 21:]
        moves = moves.view(*moves.shape[:-1], 8, 2)
        moves_id = moves[..., 0]
        moves_pp = moves[..., 1]

        entity_embed = self.pokedex_embedding(species)
        entity_embed += self.active_embedding(torch.ones_like(slot))
        entity_embed += self.sideid_embedding(sideid)
        entity_embed += self.forme_embedding(forme)
        entity_embed += self.slot_embedding(slot)
        entity_embed += self.fainted_embedding(fainted)
        entity_embed += self.gender_embedding(gender)
        entity_embed += self.embed_hp(hp)
        entity_embed += self.embed_level(level)
        entity_embed += self.embed_ability(current_and_base_ability)
        entity_embed += self.embed_item(
            curr_item, curr_item_effect, prev_item, prev_item_effect
        )
        entity_embed += self.embed_status(
            status, status_stage, sleep_turns, toxic_turns
        )
        entity_embed += self.embed_last_move(last_move)
        entity_embed += self.embed_moveset(moves_id, moves_pp)

        if self.gen == 9:
            entity_embed += self.teratype_embedding(terastallized)
            entity_embed += self.times_attacked_embedding(times_attacked)

        mask = (species == 0) | (fainted == 2)

        return entity_embed, mask

    def forward(
        self,
        n: torch.Tensor,
        total_pokemon: torch.Tensor,
        faint_counter: torch.Tensor,
        side_conditions: torch.Tensor,
        wisher: torch.Tensor,
        active: torch.Tensor,
        reserve: torch.Tensor,
        stealthrock: torch.Tensor,
        spikes: torch.Tensor,
        toxicspikes: torch.Tensor,
        stickyweb: torch.Tensor,
    ):
        scalar_emb = self.embed_scalars(
            n,
            total_pokemon,
            faint_counter,
            side_conditions,
            wisher,
            stealthrock,
            spikes,
            toxicspikes,
            stickyweb,
        )

        active_entity_embed, active_mask, boosts, volatiles = self.embed_active(active)
        reserve_entity_embed, reserve_mask = self.embed_reserve(reserve)

        entity_embeddings = torch.cat(
            [
                active_entity_embed,
                reserve_entity_embed,
            ],
            dim=-2,
        )

        T, B, S, L, *_ = entity_embeddings.shape

        mask = torch.cat(
            [
                active_mask,
                reserve_mask,
            ],
            dim=-1,
        )
        mask = mask.bool()
        mask = mask.view(T * B, S * L)

        encoder_mask = mask.clone()
        encoder_mask[..., 0] = False

        entity_embeddings: torch.Tensor = entity_embeddings.view(T * B, S * L, -1)
        entity_embeddings = self.transformer(entity_embeddings, encoder_mask)

        mask = ~mask.unsqueeze(-1)
        entity_embeddings = entity_embeddings * mask
        entity_embedding = self.output(entity_embeddings)

        entity_embeddings = entity_embeddings.view(T * B, S, L, -1)
        mask = mask.view(T * B, S, L, -1)

        spatial_embedding = self.spatial(
            entity_embeddings, mask, boosts, volatiles, scalar_emb
        )
        spatial_embedding = spatial_embedding.view(T, B, -1)

        entity_embedding = entity_embedding.view(T, B, -1)

        return entity_embedding, spatial_embedding



import json
import warnings

try:
    import torch
except ModuleNotFoundError:
    warnings.warn("torch not found")

try:
    import numpy as np
except ModuleNotFoundError:
    warnings.warn("numpy not found")

from meloetta.room import BattleRoom
from meloetta.utils import expand_bt

from typing import Union, NamedTuple, Tuple, List, Dict, Any

from meloetta.data import (
    BOOSTS,
    VOLATILES,
    PSEUDOWEATHERS,
    SIDE_CONDITIONS,
    get_item_effect_token,
    get_status_token,
    get_gender_token,
    get_species_token,
    get_ability_token,
    get_item_token,
    get_move_token,
    get_type_token,
    get_weather_token,
)


_DEFAULT_BACKEND = "torch"
_ACTIVE_SIZE = 158
_RESERVE_SIZE = 37


class NamedVector(NamedTuple):
    vector: Union[np.ndarray, torch.Tensor]
    schema: Dict[str, Tuple[int, int]]


class PublicSide(NamedTuple):
    n: torch.Tensor
    total_pokemon: torch.Tensor
    faint_counter: torch.Tensor
    side_conditions: torch.Tensor
    wisher: torch.Tensor
    active: torch.Tensor
    reserve: torch.Tensor
    stealthrock: torch.Tensor
    spikes: torch.Tensor
    toxicspikes: torch.Tensor
    stickyweb: torch.Tensor


class PrivateSide(NamedTuple):
    reserve: torch.Tensor


class PrivatePokemon(NamedTuple):
    ability: Union[torch.Tensor, None]
    active: Union[torch.Tensor, None]
    fainted: Union[torch.Tensor, None]
    gender: Union[torch.Tensor, None]
    hp: Union[torch.Tensor, None]
    item: Union[torch.Tensor, None]
    level: Union[torch.Tensor, None]
    maxhp: Union[torch.Tensor, None]
    name: Union[torch.Tensor, None]
    forme: Union[torch.Tensor, None]
    stat_atk: Union[torch.Tensor, None]
    stat_def: Union[torch.Tensor, None]
    stat_spa: Union[torch.Tensor, None]
    stat_spd: Union[torch.Tensor, None]
    stat_spe: Union[torch.Tensor, None]
    status: Union[torch.Tensor, None]
    canGmax: Union[torch.Tensor, None]
    commanding: Union[torch.Tensor, None]
    reviving: Union[torch.Tensor, None]
    teraType: Union[torch.Tensor, None]
    terastallized: Union[torch.Tensor, None]
    moves: Union[torch.Tensor, None]

    def vector(
        self,
        backend: str = _DEFAULT_BACKEND,
        with_schema: bool = False,
        *args,
        **kwargs
    ):
        arr = []
        schema = {}

        for field in self._fields:
            value = getattr(self, field)

            start = len(arr)
            if isinstance(value, int) or isinstance(value, float):
                arr.append(value)
            elif isinstance(value, list):
                arr += value

            finish = len(arr)
            schema[field] = (start, finish)

        if backend == "numpy":
            arr = np.array(arr, *args, **kwargs)
        elif backend == "torch":
            arr = torch.tensor(arr, *args, **kwargs)
        else:
            raise ValueError("Invalid Backend, must be one of `numpy` or `torch`")

        if with_schema:
            return NamedVector(arr, schema)
        else:
            return arr


class ReservePublicPokemon(NamedTuple):
    species: torch.Tensor
    forme: torch.Tensor
    slot: torch.Tensor
    hp: torch.Tensor
    fainted: torch.Tensor
    level: torch.Tensor
    gender: torch.Tensor
    ability: torch.Tensor
    base_ability: torch.Tensor
    item: torch.Tensor
    item_effect: torch.Tensor
    prev_item: torch.Tensor
    prev_item_effect: torch.Tensor
    terastallized: torch.Tensor
    status: torch.Tensor
    status_stage: torch.Tensor
    last_move: torch.Tensor
    times_attacked: torch.Tensor
    sleep_turns: torch.Tensor
    toxic_turns: torch.Tensor
    sideid: torch.Tensor
    moves: torch.Tensor

    def vector(
        self,
        backend: str = _DEFAULT_BACKEND,
        with_schema: bool = False,
        *args,
        **kwargs
    ):
        arr = []
        schema = {}

        for field in self._fields:
            value = getattr(self, field)

            start = len(arr)
            if isinstance(value, int) or isinstance(value, float):
                arr.append(value)
            elif isinstance(value, list):
                arr += value
            finish = len(arr)
            schema[field] = (start, finish)

        if backend == "numpy":
            arr = np.array(arr, *args, **kwargs)
        elif backend == "torch":
            arr = torch.tensor(arr, *args, **kwargs)
        else:
            raise ValueError("Invalid Backend, must be one of `numpy` or `torch`")

        if with_schema:
            return NamedVector(arr, schema)
        else:
            return arr


class ActivePublicPokemon(NamedTuple):
    species: torch.Tensor
    forme: torch.Tensor
    slot: torch.Tensor
    hp: torch.Tensor
    fainted: torch.Tensor
    level: torch.Tensor
    gender: torch.Tensor
    ability: torch.Tensor
    base_ability: torch.Tensor
    item: torch.Tensor
    item_effect: torch.Tensor
    prev_item: torch.Tensor
    prev_item_effect: torch.Tensor
    terastallized: torch.Tensor
    status: torch.Tensor
    status_stage: torch.Tensor
    last_move: torch.Tensor
    times_attacked: torch.Tensor
    sleep_turns: torch.Tensor
    toxic_turns: torch.Tensor
    boosts: torch.Tensor
    volatiles: torch.Tensor
    side: torch.Tensor
    moves: torch.Tensor

    def vector(
        self,
        backend: str = _DEFAULT_BACKEND,
        with_schema: bool = False,
        *args,
        **kwargs
    ):
        arr = []
        schema = {}

        for field in self._fields:
            value = getattr(self, field)

            start = len(arr)
            if isinstance(value, int) or isinstance(value, float):
                arr.append(value)
            elif isinstance(value, list):
                arr += value
            finish = len(arr)
            schema[field] = (start, finish)

        if backend == "numpy":
            arr = np.array(arr, *args, **kwargs)
        elif backend == "torch":
            arr = torch.tensor(arr, *args, **kwargs)
        else:
            raise ValueError("Invalid Backend, must be one of `numpy` or `torch`")

        if with_schema:
            return NamedVector(arr, schema)
        else:
            return arr


class State(NamedTuple):
    player_id: torch.Tensor
    private_side: PrivateSide
    public_sides: PublicSide
    weather: torch.Tensor
    weather_time_left: torch.Tensor
    weather_min_time_left: torch.Tensor
    pseudo_weather: torch.Tensor
    turn: torch.Tensor
    log: List[str]

    def to_dict(self):
        return {
            "player_id": self.player_id,
            "private_reserve": self.private_side.reserve,
            **{
                "public_" + key: value
                for key, value in self.public_sides._asdict().items()
            },
            "weather": self.weather,
            "weather_time_left": self.weather_time_left,
            "weather_min_time_left": self.weather_min_time_left,
            "pseudo_weather": self.pseudo_weather,
            "turn": self.turn,
        }


Battle = Dict[str, Dict[str, Dict[str, Any]]]


LAST_MOVES = set()


def hashify_pokemon(pokemon: Dict[str, Any]):
    info = pokemon["name"]
    info += pokemon["ident"]
    info += pokemon["speciesForme"]
    info += pokemon["details"]
    info += pokemon["searchid"]
    return hash(info)


class VectorizedState:
    def __init__(self, room: BattleRoom, battle: Battle):
        self.room = room
        self.battle = battle
        self.gen = self.battle["dex"]["gen"]
        self.moves = {}

    @classmethod
    def from_battle(self, room: BattleRoom, battle: Battle) -> State:
        vstate = VectorizedState(room, battle)
        return vstate.vectorize()

    def vectorize(self) -> State:
        public_sides = self._vectorize_public_sides()
        private_side = self._vectorize_private_side()

        sideid = self.battle["mySide"]["sideid"]
        player_id = 0 if sideid == "p1" else 1
        player_id = torch.tensor(player_id)
        player_id = expand_bt(player_id)

        pseudoweathers = {
            (pseudoweather[0]).replace(" ", "").lower(): pseudoweather[1:]
            for pseudoweather in self.battle["pseudoWeather"]
        }
        pseudoweathers = torch.stack(
            [
                torch.tensor(pseudoweathers.get(pseudoweather, [-1, -1]))
                for pseudoweather in PSEUDOWEATHERS
            ]
        )
        pseudoweathers = expand_bt(pseudoweathers)

        weather = torch.tensor(get_weather_token(self.battle["weather"]))
        weather = expand_bt(weather)

        weather_time_left = torch.tensor(self.battle["weatherTimeLeft"])
        weather_time_left = expand_bt(weather_time_left)

        weather_min_time_left = torch.tensor(self.battle["weatherMinTimeLeft"])
        weather_min_time_left = expand_bt(weather_min_time_left)

        turn = torch.tensor(self.battle["turn"])
        turn = expand_bt(turn)

        return State(
            player_id=player_id,
            private_side=private_side,
            public_sides=public_sides,
            weather=weather,
            weather_time_left=weather_time_left,
            weather_min_time_left=weather_min_time_left,
            turn=turn,
            pseudo_weather=pseudoweathers,
            log=self.battle.get("stepQueue", []),
        )

    def _vectorize_public_sides(self) -> PublicSide:
        p1 = self._vectorize_public_side("mySide")
        p2 = self._vectorize_public_side("farSide")
        combined_fields = {}
        for field, value1, value2 in zip(PublicSide._fields, p1, p2):
            combined_fields[field] = torch.stack((value1, value2), dim=2)
        return PublicSide(**combined_fields)

    def _vectorize_private_side(self) -> PrivateSide:
        reserve = [
            self._vectorize_private_pokemon(pokemon)
            for pokemon in (self.battle["myPokemon"])
        ]
        reserve += [-torch.ones_like(reserve[0]) for _ in range(6 - len(reserve))]

        reserve = torch.stack(reserve)
        reserve = expand_bt(reserve)
        return PrivateSide(reserve=reserve)

    def _vectorize_private_pokemon(self, pokemon: Dict[str, Any]):
        if self.gen == 9:
            if pokemon.get("commanding"):
                commanding = 1
            else:
                commanding = 0

            if pokemon.get("reviving"):
                reviving = 1
            else:
                reviving = 0

            if pokemon.get("teraType"):
                teraType = get_type_token(self.gen, pokemon["teraType"])
            else:
                teraType = -1

            if pokemon.get("terastallized"):
                terastallized = 1
            else:
                terastallized = 0

        else:
            commanding = None
            reviving = None
            teraType = None
            terastallized = None

        if self.gen == 8:
            if pokemon.get("canGmax"):
                canGmax = 1
            else:
                canGmax = 0
        else:
            canGmax = None

        myside = self.moves.get("mySide") or {}
        move_track = myside.get(pokemon["ident"]) or {}

        moves = []
        for move in pokemon["moves"]:
            pp = move_track.get(move, 0)
            moves += [get_move_token(self.gen, "id", move), pp]

        for _ in range(4 - int(len(moves) / 2)):
            moves += [-1, -1]

        forme = pokemon["speciesForme"].replace(pokemon["name"] + "-", "")

        private_pokemon = PrivatePokemon(
            ability=get_ability_token(self.gen, "id", pokemon["baseAbility"]),
            active=1 if pokemon["active"] else 0,
            fainted=1 if pokemon.get("fainted") else 0,
            gender=get_gender_token(pokemon["gender"]),
            hp=pokemon["hp"],
            item=get_item_token(self.gen, "id", pokemon["item"]),
            level=pokemon["level"],
            maxhp=pokemon["maxhp"],
            name=get_species_token(self.gen, "name", pokemon["name"]),
            forme=get_species_token(self.gen, "forme", forme),
            stat_atk=pokemon["stats"]["atk"],
            stat_def=pokemon["stats"]["def"],
            stat_spa=pokemon["stats"]["spa"],
            stat_spd=pokemon["stats"]["spd"],
            stat_spe=pokemon["stats"]["spe"],
            status=get_status_token(pokemon.get("status")),
            canGmax=canGmax,
            commanding=commanding,
            reviving=reviving,
            teraType=teraType,
            terastallized=terastallized,
            moves=moves,
        )
        return private_pokemon.vector()

    def _vectorize_public_side(self, side_id: str):
        side = self.battle[side_id]

        controlling = self.battle["pokemonControlled"]

        # This is a measure for removing duplicates from zoroak
        # To be clear, this doesn't remove zoroark, but only mons
        # that are affected by -replace
        # There could be up to 5 extra mons generated from zoroark alone
        # TODO: account for zororak duplicates somehow...
        # active_pokemon = {}
        # for pokemon in reversed(side["active"]):
        #     if pokemon is not None:
        #         active_pokemon[hashify_pokemon(pokemon)] = pokemon

        # reserve_pokemon = {}
        # for pokemon in reversed(side["pokemon"]):
        #     if pokemon is not None:
        #         pokemon_hash = hashify_pokemon(pokemon)
        #         if pokemon_hash not in active_pokemon:
        #             reserve_pokemon[pokemon_hash] = pokemon

        # active_pokemon = list(active_pokemon.values())
        # reserve_pokemon = list(reserve_pokemon.values())[:6]

        pokemon = {p["ident"]: p for p in reversed(side["pokemon"])}
        pokemon = list(reversed(pokemon.values()))

        active_pokemon = [p for p in side["active"] if p is not None]
        active_idents = [p["ident"] for p in active_pokemon]
        reserve_pokemon = [p for p in pokemon if p["ident"] not in active_idents]

        active = [
            self._vectorize_public_active_pokemon(side_id, p) for p in active_pokemon
        ]
        active += [-torch.ones(_ACTIVE_SIZE) for _ in range(controlling - len(active))]
        active = torch.stack(active)
        active = expand_bt(active)

        reserve = [
            self._vectorize_public_reserve_pokemon(side_id, p) for p in reserve_pokemon
        ]
        num_reserve_padding = 6 - len(reserve)
        reserve += [-torch.ones(_RESERVE_SIZE) for _ in range(num_reserve_padding)]

        reserve = torch.stack(reserve)
        reserve = expand_bt(reserve)

        side_conditions = torch.stack(
            [
                torch.tensor(
                    side["sideConditions"].get(side_condition, [None, 0, -1, -1])[1:]
                )
                for side_condition in SIDE_CONDITIONS
                if side_condition
                not in {"stealthrock", "spikes", "toxicspikes", "stickyweb"}
            ]
        )
        # [effectName, levels, minDuration, maxDuration]
        side_conditions = expand_bt(side_conditions)

        stealthrock = torch.tensor(
            side["sideConditions"].get("stealthrock", [None, 0])[1]
        )
        stealthrock = expand_bt(stealthrock)
        spikes = torch.tensor(side["sideConditions"].get("spikes", [None, 0])[1])
        spikes = expand_bt(spikes)
        toxicspikes = torch.tensor(
            side["sideConditions"].get("toxicspikes", [None, 0])[1]
        )
        toxicspikes = expand_bt(toxicspikes)
        stickyweb = torch.tensor(side["sideConditions"].get("stickyweb", [None, 0])[1])
        stickyweb = expand_bt(stickyweb)

        n = torch.tensor(side["n"])
        n = expand_bt(n)

        total_pokemon = torch.tensor(side["totalPokemon"])
        total_pokemon = expand_bt(total_pokemon)

        faint_counter = torch.tensor(min(6, side["faintCounter"]))
        faint_counter = expand_bt(faint_counter)

        wisher_slot = (side["wisher"] or {"slot": -1})["slot"]
        wisher_slot = torch.tensor(wisher_slot)
        wisher_slot = expand_bt(wisher_slot)

        return PublicSide(
            n=n,
            total_pokemon=total_pokemon,
            faint_counter=faint_counter,
            active=active,
            reserve=reserve,
            side_conditions=side_conditions,
            stealthrock=stealthrock,
            spikes=spikes,
            toxicspikes=toxicspikes,
            stickyweb=stickyweb,
            wisher=wisher_slot,
        )

    def _vectorize_public_active_pokemon(self, side_id: str, pokemon: Dict[str, Any]):
        if pokemon is None:
            return None

        moves = []
        if side_id not in self.moves:
            self.moves[side_id] = {}
        if pokemon["ident"] not in self.moves[side_id]:
            self.moves[side_id][pokemon["ident"]] = {}

        for move, pp in pokemon["moveTrack"]:
            pp = max(pp, 0)
            self.moves[side_id][pokemon["ident"]][move] = pp
            moves += [get_move_token(self.gen, "name", move), pp]

        for _ in range(8 - int(len(moves) / 2)):
            moves += [-1, -1]

        boosts = [pokemon["boosts"].get(boost, 0) for boost in BOOSTS]
        volatiles = [
            1 if volatile in pokemon["volatiles"] else 0 for volatile in VOLATILES
        ]
        forme = pokemon["speciesForme"].replace(pokemon["name"] + "-", "")
        active = ActivePublicPokemon(
            species=get_species_token(self.gen, "name", pokemon["name"]),
            forme=get_species_token(self.gen, "forme", forme),
            slot=pokemon["slot"],
            hp=pokemon["hp"] / pokemon["maxhp"],
            fainted=1 if pokemon["fainted"] else 0,
            level=pokemon["level"],
            gender=get_gender_token(pokemon["gender"]),
            ability=get_ability_token(self.gen, "name", pokemon["ability"]),
            base_ability=get_ability_token(self.gen, "name", pokemon["baseAbility"]),
            item=get_item_token(self.gen, "name", pokemon["item"]),
            item_effect=get_item_effect_token(pokemon["itemEffect"]),
            prev_item=get_item_token(self.gen, "name", pokemon["prevItem"]),
            prev_item_effect=get_item_effect_token(pokemon["prevItemEffect"]),
            terastallized=get_type_token(self.gen, pokemon["terastallized"]),
            status=get_status_token(pokemon["status"]),
            status_stage=pokemon["statusStage"],
            last_move=get_move_token(self.gen, "id", pokemon["lastMove"]),
            times_attacked=min(pokemon["timesAttacked"], 6),
            sleep_turns=min(pokemon["statusData"]["sleepTurns"], 15),
            toxic_turns=min(pokemon["statusData"]["toxicTurns"], 3),
            boosts=boosts,
            volatiles=volatiles,
            side=0 if side_id == "mySide" else 1,
            moves=moves,
        )
        return active.vector()

    def _vectorize_public_reserve_pokemon(self, side_id: str, pokemon: Dict[str, Any]):
        moves = []
        for move, pp in pokemon["moveTrack"]:
            moves += [get_move_token(self.gen, "name", move), pp]

        self.moves[pokemon["ident"]] = moves
        for _ in range(8 - int(len(moves) / 2)):
            moves += [-1, -1]

        forme = pokemon["speciesForme"].replace(pokemon["name"] + "-", "")

        reserve = ReservePublicPokemon(
            species=get_species_token(self.gen, "name", pokemon["name"]),
            forme=get_species_token(self.gen, "forme", forme),
            slot=pokemon["slot"],
            hp=pokemon["hp"] / pokemon["maxhp"],
            fainted=1 if pokemon["fainted"] else 0,
            level=pokemon["level"],
            gender=get_gender_token(pokemon["gender"]),
            ability=get_ability_token(self.gen, "name", pokemon["ability"]),
            base_ability=get_ability_token(self.gen, "name", pokemon["baseAbility"]),
            item=get_item_token(self.gen, "name", pokemon["item"]),
            item_effect=get_item_effect_token(pokemon["itemEffect"]),
            prev_item=get_item_token(self.gen, "name", pokemon["prevItem"]),
            prev_item_effect=get_item_effect_token(pokemon["prevItemEffect"]),
            terastallized=get_type_token(self.gen, pokemon["terastallized"]),
            status=get_status_token(pokemon["status"]),
            status_stage=pokemon["statusStage"],
            last_move=get_move_token(self.gen, "name", pokemon["lastMove"]),
            times_attacked=min(pokemon["timesAttacked"], 6),
            sleep_turns=min(pokemon["statusData"]["sleepTurns"], 15),
            toxic_turns=min(pokemon["statusData"]["toxicTurns"], 3),
            sideid=0 if side_id == "mySide" else 1,
            moves=moves,  #
        )
        return reserve.vector()



import json

from dataclasses import dataclass

from typing import Any, Sequence

from meloetta.frameworks.mewzero.model import config
from meloetta.frameworks.mewzero import utils


def is_jsonable(x):
    try:
        json.dumps(x)
        return True
    except:
        return False


@dataclass
class AdamConfig:
    """Adam optimizer related params."""

    b1: float = 0.0
    b2: float = 0.999
    eps: float = 10e-8


@dataclass
class NerdConfig:
    """Nerd related params."""

    beta: float = 2.0
    clip: float = 10_000


@dataclass
class MewZeroConfig:
    """Learning pararms"""

    # The batch size to use when learning/improving parameters.
    batch_size: int = 32
    # The learning rate for `params`.
    learning_rate: float = 5e-5
    # The config related to the ADAM optimizer used for updating `params`.
    adam: AdamConfig = AdamConfig()
    # All gradients values are clipped to [-clip_gradient, clip_gradient].
    clip_gradient: float = 10_000
    # The "speed" at which `params_target` is following `params`.
    target_network_avg: float = 0.01

    # RNaD algorithm configuration.
    # Entropy schedule configuration. See EntropySchedule class documentation.
    entropy_schedule_repeats: Sequence[int] = (1,)
    entropy_schedule_size: Sequence[int] = (50,)
    # The weight of the reward regularisation term in RNaD.
    eta_reward_transform: float = 0.2
    gamma: float = 1.0
    nerd: NerdConfig = NerdConfig()
    c_vtrace: float = 1.0

    trajectory_length: int = 512

    # battle_format = "gen8randomdoublesbattle"
    # battle_format = "gen8randombattle"
    battle_format: str = "gen9randombattle"
    # battle_format = "gen8ou"
    # battle_format = "gen9ou"
    # battle_format = "gen8doublesou"
    # battle_format: str = "gen3randombattle"
    # battle_format = "gen9doublesou"

    gen, format = utils.get_gen_and_gametype(battle_format)

    team: str = "null"
    # team = "charizard||heavydutyboots|blaze|furyswipes,scaleshot,toxic,roost||85,,85,85,85,85||,0,,,,||88|"
    # team = "charizard||heavydutyboots|blaze|hurricane,fireblast,toxic,roost||85,,85,85,85,85||,0,,,,||88|]venusaur||blacksludge|chlorophyll|leechseed,substitute,sleeppowder,sludgebomb||85,,85,85,85,85||,0,,,,||82|]blastoise||whiteherb|torrent|shellsmash,earthquake,icebeam,hydropump||85,85,85,85,85,85||||86|"
    # team = "charizard||heavydutyboots|blaze|hurricane,fireblast,toxic,roost||85,,85,85,85,85||,0,,,,||88|]blastoise||whiteherb|torrent|shellsmash,earthquake,icebeam,hydropump||85,85,85,85,85,85||||86|"
    # team = "ceruledge||lifeorb|weakarmor|bitterblade,closecombat,shadowsneak,swordsdance||85,85,85,85,85,85||||82|,,,,,fighting]grafaiai||leftovers|prankster|encore,gunkshot,knockoff,partingshot||85,85,85,85,85,85||||86|,,,,,dark]greedent||sitrusberry|cheekpouch|bodyslam,psychicfangs,swordsdance,firefang||85,85,85,85,85,85||||88|,,,,,psychic]quaquaval||lifeorb|moxie|aquastep,closecombat,swordsdance,icespinner||85,85,85,85,85,85||||80|,,,,,fighting]flapple||lifeorb|hustle|gravapple,outrage,dragondance,suckerpunch||85,85,85,85,85,85||||84|,,,,,grass]pachirisu||assaultvest|voltabsorb|nuzzle,superfang,thunderbolt,uturn||85,85,85,85,85,85||||94|,,,,,flying"

    actor_device: str = "cpu"
    learner_device: str = "cuda"

    debug_mode = False

    # This config will spawn 20 workers with 2 players each
    # for a total of 40 players, playing 20 games.
    # it is recommended to have an even number of players per worker
    num_actors: int = 1 if debug_mode else 12
    num_buffers: int = max(4 * num_actors, 2 * batch_size)

    model_config: config.MewZeroModelConfig = config.MewZeroModelConfig()

    eval: bool = not debug_mode

    def __getindex__(self, key: str):
        return self.__dict__[key]

    def __repr__(self):
        d = {
            key: value if is_jsonable(value) else repr(value)
            for key, value in self.__dict__.items()
        }
        body = json.dumps(d, indent=4, sort_keys=True)
        return f"MewZeroConfig({body})"

    def get(self, key: Any, default=None):
        return self.__dict__.get(key, default)



import torch

from typing import Tuple

from meloetta.room import BattleRoom

from meloetta.frameworks.mewzero import ReplayBuffer
from meloetta.frameworks.mewzero.model import (
    MewZeroModel,
    EnvStep,
    PostProcess,
)

from meloetta.actors.base import Actor
from meloetta.actors.types import State, Choices


class MewZeroActor(Actor):
    def __init__(
        self,
        model: MewZeroModel,
        replay_buffer: ReplayBuffer = None,
    ):
        self.model = model
        self.replay_buffer = replay_buffer
        self.hidden_state = model.core.initial_state(1)

    @property
    def storing_transition(self):
        return self.replay_buffer is not None

    def choose_action(
        self,
        state: State,
        room: BattleRoom,
        choices: Choices,
    ):
        output: Tuple[EnvStep, PostProcess]
        with torch.no_grad():
            output = self.model(state, self.hidden_state, choices)
        env_step, postprocess, self.hidden_state = output

        if self.storing_transition:
            self.store_transition(state, env_step, room)

        data = postprocess.data
        index = postprocess.index
        func, args, kwargs = data[index.item()]
        return func, args, kwargs

    def store_transition(self, state: State, env_step: EnvStep, room: BattleRoom):
        state = self.model.clean(state)
        to_store = env_step.to_store(state)
        self.replay_buffer.store_sample(room.battle_tag, to_store)

    def post_match(self, room: BattleRoom):
        if self.storing_transition:
            datum = room.get_reward()
            self.replay_buffer.append_reward(
                room.battle_tag, datum["pid"], datum["reward"]
            )
            self.replay_buffer.register_done(room.battle_tag)



import torch.nn as nn

from meloetta.frameworks.nash_ketchum.model.interfaces import (
    State,
    EncoderOutput,
    SideEncoderOutput,
)
from meloetta.frameworks.nash_ketchum.model.config import EncoderConfig
from meloetta.frameworks.nash_ketchum.model.encoders import (
    SideEncoder,
    ScalarEncoder,
    WeatherEncoder,
)


class Encoder(nn.Module):
    def __init__(self, gen: int, n_active: int, config: EncoderConfig):
        super().__init__()
        self.side_encoder = SideEncoder(
            gen=gen, n_active=n_active, config=config.side_encoder_config
        )
        self.weather_encoder = WeatherEncoder(
            gen=gen, config=config.weather_encoder_config
        )
        self.scalar_encoder = ScalarEncoder(
            gen=gen, n_active=n_active, config=config.scalar_encoder_config
        )

    def forward(self, state: State) -> EncoderOutput:
        sides = state["sides"]
        boosts = state["boosts"]
        volatiles = state["volatiles"]
        side_conditions = state["side_conditions"]
        pseudoweathers = state["pseudoweathers"]
        weather = state["weather"]
        wisher = state["wisher"]

        # scalars
        scalars = state["scalars"]
        turn = scalars[..., 0]
        n = scalars[..., 1:3]
        total_pokemon = scalars[..., 3:5]
        faint_counter = scalars[..., 5:]

        # action masks
        action_type_mask = state["action_type_mask"]
        move_mask = state["move_mask"]
        switch_mask = state["switch_mask"]
        flag_mask = state["flag_mask"]
        max_move_mask = state.get("max_move_mask")
        target_mask = state.get("target_mask")

        side_embs = self.side_encoder(
            side=sides,
            boosts=boosts,
            volatiles=volatiles,
            side_conditions=side_conditions,
            wisher=wisher,
        )

        weather_emb = self.weather_encoder(
            weather=weather,
            pseudoweather=pseudoweathers,
        )

        scalar_emb = self.scalar_encoder(
            turn=turn,
            action_type_mask=action_type_mask,
            move_mask=move_mask,
            max_move_mask=max_move_mask,
            switch_mask=switch_mask,
            flag_mask=flag_mask,
            target_mask=target_mask,
            n=n,
            total_pokemon=total_pokemon,
            faint_counter=faint_counter,
        )

        return EncoderOutput(
            **side_embs._asdict(),
            weather_emb=weather_emb,
            scalar_emb=scalar_emb,
        )



import re
import torch
import collections
import torch.nn.functional as F

from typing import Dict, List, Tuple

from meloetta.frameworks.porygon.model.utils import _log_policy, _legal_policy
from meloetta.data import CHOICE_FLAGS


def get_gen_and_gametype(battle_format: str) -> Tuple[int, str]:
    gen = int(re.search(r"gen([0-9])", battle_format).groups()[0])
    if "triples" in battle_format:
        gametype = "triples"
    if "doubles" in battle_format:
        gametype = "doubles"
    else:
        gametype = "singles"
    return gen, gametype


def to_id(string: str):
    return "".join(c for c in string if c.isalnum()).lower()


def _get_private_reserve_size(gen: int):
    if gen == 9:
        return 28
    elif gen == 8:
        return 25
    else:
        return 24


def get_buffer_specs(
    trajectory_length: int,
    gen: int,
    gametype: str,
    private_reserve_size: int,
):
    if gametype != "singles":
        if gametype == "doubles":
            n_active = 2
        else:
            n_active = 3
    else:
        n_active = 1

    buffer_specs = {
        "sides": {
            "size": (trajectory_length, 3, 12, 32),
            "dtype": torch.int64,
        },
        "boosts": {
            "size": (trajectory_length, 2, 8),
            "dtype": torch.int64,
        },
        "volatiles": {
            "size": (trajectory_length, 2, 113),
            "dtype": torch.int64,
        },
        "side_conditions": {
            "size": (trajectory_length, 2, 58),
            "dtype": torch.int64,
        },
        "pseudoweathers": {
            "size": (trajectory_length, 12, 2),
            "dtype": torch.int64,
        },
        "weather": {
            "size": (trajectory_length, 3),
            "dtype": torch.int64,
        },
        "wisher": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "turn": {
            "size": (trajectory_length,),
            "dtype": torch.int64,
        },
        "n": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "total_pokemon": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "faint_counter": {
            "size": (trajectory_length, 2),
            "dtype": torch.int64,
        },
        "turns_since_last_move": {
            "size": (trajectory_length,),
            "dtype": torch.int64,
        },
        "action_type_mask": {
            "size": (trajectory_length, 3),
            "dtype": torch.bool,
        },
        "move_mask": {
            "size": (trajectory_length, 4),
            "dtype": torch.bool,
        },
        "switch_mask": {
            "size": (trajectory_length, 6),
            "dtype": torch.bool,
        },
        "flag_mask": {
            "size": (trajectory_length, len(CHOICE_FLAGS)),
            "dtype": torch.bool,
        },
        "rewards": {
            "size": (trajectory_length,),
            "dtype": torch.float32,
        },
        "player_id": {
            "size": (trajectory_length,),
            "dtype": torch.long,
        },
    }
    if gametype != "singles":
        buffer_specs.update(
            {
                "target_mask": {
                    "size": (trajectory_length, 4),
                    "dtype": torch.bool,
                },
                "prev_choices": {
                    "size": (trajectory_length, 2, 4),
                    "dtype": torch.int64,
                },
                "choices_done": {
                    "size": (trajectory_length, 1),
                    "dtype": torch.int64,
                },
                "targeting": {
                    "size": (trajectory_length, 1),
                    "dtype": torch.int64,
                },
                "target_policy": {
                    "size": (trajectory_length, 2 * n_active),
                    "dtype": torch.float32,
                },
                "target_index": {
                    "size": (trajectory_length,),
                    "dtype": torch.int64,
                },
            }
        )

    # add policy...
    buffer_specs.update(
        {
            "action_type_policy": {
                "size": (trajectory_length, 3),
                "dtype": torch.float32,
            },
            "move_policy": {
                "size": (trajectory_length, 4),
                "dtype": torch.float32,
            },
            "switch_policy": {
                "size": (trajectory_length, 6),
                "dtype": torch.float32,
            },
            "flag_policy": {
                "size": (trajectory_length, len(CHOICE_FLAGS)),
                "dtype": torch.float32,
            },
        }
    )

    # ...and indices
    buffer_specs.update(
        {
            "action_type_index": {
                "size": (trajectory_length,),
                "dtype": torch.int64,
            },
            "move_index": {
                "size": (trajectory_length,),
                "dtype": torch.int64,
            },
            "switch_index": {
                "size": (trajectory_length,),
                "dtype": torch.int64,
            },
            "flag_index": {
                "size": (trajectory_length,),
                "dtype": torch.int64,
            },
        }
    )

    if gen == 8:
        buffer_specs.update(
            {
                "max_move_mask": {
                    "size": (trajectory_length, 4),
                    "dtype": torch.bool,
                },
                "max_move_policy": {
                    "size": (trajectory_length, 4),
                    "dtype": torch.float32,
                },
                "max_move_index": {
                    "size": (trajectory_length,),
                    "dtype": torch.int64,
                },
            }
        )

    buffer_specs.update(
        {
            "valid": {
                "size": (trajectory_length,),
                "dtype": torch.bool,
            },
            "value": {
                "size": (trajectory_length,),
                "dtype": torch.float32,
            },
        },
    )
    return buffer_specs


def create_buffers(num_buffers: int, trajectory_length: int, gen: int, gametype: str):
    """
    num_buffers: int
        the size of the replay buffer
    trajectory_length: int
        the max length of a replay
    battle_format: str
        the type of format
    """

    private_reserve_size = _get_private_reserve_size(gen)

    buffer_specs = get_buffer_specs(
        trajectory_length, gen, gametype, private_reserve_size
    )

    buffers: Dict[str, List[torch.Tensor]] = {key: [] for key in buffer_specs}
    for _ in range(num_buffers):
        for key in buffer_specs:
            if key.endswith("_mask"):
                buffers[key].append(torch.ones(**buffer_specs[key]).share_memory_())
            else:
                buffers[key].append(torch.zeros(**buffer_specs[key]).share_memory_())
    return buffers


def expand_bt(tensor: torch.Tensor, time: int = 1, batch: int = 1) -> torch.Tensor:
    shape = tensor.shape
    return tensor.view(time, batch, *(tensor.shape or shape))


VTraceFromLogitsReturns = collections.namedtuple(
    "VTraceFromLogitsReturns",
    [
        "vs",
        "pg_advantages",
        "log_rhos",
        "behavior_action_log_probs",
        "target_action_log_probs",
        "kl_loss",
    ],
)

VTraceReturns = collections.namedtuple("VTraceReturns", "vs pg_advantages")


def action_log_probs(policy_logits, actions, policy_mask):
    return torch.gather(
        _log_policy(policy_logits, policy_mask), -1, actions.unsqueeze(-1)
    ).view_as(actions)


def from_logits(
    behavior_policy_logits,
    target_policy_logits,
    policy_mask,
    actions,
    discounts,
    rewards,
    values,
    bootstrap_value,
    clip_rho_threshold=1.0,
    clip_pg_rho_threshold=1.0,
):
    """V-trace for softmax policies."""

    target_action_log_probs = action_log_probs(
        target_policy_logits, actions, policy_mask
    )
    behavior_action_log_probs = action_log_probs(
        behavior_policy_logits, actions, policy_mask
    )
    log_rhos = target_action_log_probs - behavior_action_log_probs
    vtrace_returns = from_importance_weights(
        log_rhos=log_rhos,
        discounts=discounts,
        rewards=rewards,
        values=values,
        bootstrap_value=bootstrap_value,
        clip_rho_threshold=clip_rho_threshold,
        clip_pg_rho_threshold=clip_pg_rho_threshold,
    )

    target_dist = torch.distributions.Categorical(
        probs=_legal_policy(target_policy_logits, policy_mask)
    )
    behavior_dist = torch.distributions.Categorical(
        probs=_legal_policy(behavior_policy_logits, policy_mask)
    )
    kl_loss = torch.distributions.kl.kl_divergence(target_dist, behavior_dist)

    return VTraceFromLogitsReturns(
        kl_loss=kl_loss,
        log_rhos=log_rhos,
        behavior_action_log_probs=behavior_action_log_probs,
        target_action_log_probs=target_action_log_probs,
        **vtrace_returns._asdict(),
    )


@torch.no_grad()
def from_importance_weights(
    log_rhos,
    discounts,
    rewards,
    values,
    bootstrap_value,
    clip_rho_threshold=1.0,
    clip_pg_rho_threshold=1.0,
):
    """V-trace from log importance weights."""
    with torch.no_grad():
        rhos = torch.exp(log_rhos)
        if clip_rho_threshold is not None:
            clipped_rhos = torch.clamp(rhos, max=clip_rho_threshold)
        else:
            clipped_rhos = rhos

        cs = torch.clamp(rhos, max=1.0)
        # Append bootstrapped value to get [v1, ..., v_t+1]
        values_t_plus_1 = torch.cat(
            [values[1:], torch.unsqueeze(bootstrap_value, 0)], dim=0
        )
        deltas = clipped_rhos * (rewards + discounts * values_t_plus_1 - values)

        acc = torch.zeros_like(bootstrap_value)
        result = []
        for t in range(discounts.shape[0] - 1, -1, -1):
            acc = deltas[t] + discounts[t] * cs[t] * acc
            result.append(acc)
        result.reverse()
        vs_minus_v_xs = torch.stack(result)

        # Add V(x_s) to get v_s.
        vs = torch.add(vs_minus_v_xs, values)

        # Advantage for policy gradient.
        broadcasted_bootstrap_values = torch.ones_like(vs[0]) * bootstrap_value
        vs_t_plus_1 = torch.cat(
            [vs[1:], broadcasted_bootstrap_values.unsqueeze(0)], dim=0
        )
        if clip_pg_rho_threshold is not None:
            clipped_pg_rhos = torch.clamp(rhos, max=clip_pg_rho_threshold)
        else:
            clipped_pg_rhos = rhos
        pg_advantages = clipped_pg_rhos * (rewards + discounts * vs_t_plus_1 - values)

        # Make sure no gradients backpropagated through the returned values.
        return VTraceReturns(vs=vs, pg_advantages=pg_advantages)



import json
import torch

import pandas as pd
import plotly.express as px

from transformers import AutoTokenizer, AutoModel

from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans

from tqdm import tqdm

descriptors = [
    "meloetta/js/data/BattlePokedex.json",
    "meloetta/js/data/BattleAbilities.json",
    "meloetta/js/data/BattleItems.json",
    "meloetta/js/data/BattleMovedex.json",
]


def main():

    model_id = "princeton-nlp/sup-simcse-roberta-large"
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    tokenizer = AutoTokenizer.from_pretrained(model_id)
    model = AutoModel.from_pretrained(model_id)
    model = model.to(device)

    output_dim = 3
    pca = PCA(output_dim)
    kmeans = KMeans(25)
    scaler = StandardScaler()

    for file in descriptors:
        with open(file, "r") as f:
            data = json.load(f)

        texts = [json.dumps(a) for a in sorted(data.values(), key=lambda x: x["num"])]

        outputs = []
        with torch.no_grad():
            for text in tqdm(texts):
                encoded = tokenizer(
                    text, padding=True, truncation=True, return_tensors="pt"
                )
                for key in encoded:
                    if isinstance(encoded[key], torch.Tensor):
                        encoded[key] = encoded[key].to(device)
                outputs.append(model(**encoded).pooler_output.squeeze())

        vectors = torch.stack(outputs).cpu().numpy()
        labels = kmeans.fit_predict(vectors).astype(str)
        vectors = pca.fit_transform(vectors)
        vectors = scaler.fit_transform(vectors)

        df = [a for a in sorted(data.values(), key=lambda x: x["num"])]
        for i, s in enumerate(df):
            x, y, z = vectors[i]
            s["x"] = x
            s["y"] = y
            s["z"] = z
            s["cluster"] = labels[i]

        df = pd.DataFrame(df)
        fig = px.scatter_3d(
            df,
            x="x",
            y="y",
            z="z",
            color="cluster",
            hover_data=["name"] + (["desc"] if "desc" in df.columns else []),
        )
        fig.show()

        name = file.split("/")[-1].split(".")[0]
        print(
            f"{name} - explained variance: {100 * sum(pca.explained_variance_ratio_):.2f}"
        )

        torch.save(vectors, f"pretrained/{name}.pt")


if __name__ == "__main__":
    main()



from meloetta.frameworks.nash_ketchum.model.heads.action_type_head import ActionTypeHead
from meloetta.frameworks.nash_ketchum.model.heads.move_head import MoveHead
from meloetta.frameworks.nash_ketchum.model.heads.switch_head import SwitchHead
from meloetta.frameworks.nash_ketchum.model.heads.flag_head import FlagsHead



from meloetta.frameworks.mewzero.buffer import ReplayBuffer
from meloetta.frameworks.mewzero.actor import MewZeroActor
from meloetta.frameworks.mewzero.model import MewZeroModel
from meloetta.frameworks.mewzero.learner import MewZeroLearner
from meloetta.frameworks.mewzero.config import MewZeroConfig
from meloetta.frameworks.mewzero import utils



import random

import torch
import torch.nn as nn

from meloetta.actors.types import State, Choices
from meloetta.embeddings import MoveEmbedding


class MaxDamageModel(nn.Module):
    def __init__(self, gen: int):
        super().__init__()
        self.move_embedding = MoveEmbedding(gen=gen)

    def forward(
        self,
        state: State,
        choices: Choices,
    ):
        private_reserve = state["private_reserve"].squeeze()
        private_reserve_x = private_reserve + 1

        active = private_reserve_x[..., 1] == 2

        try:
            if choices["moves"] and active.sum():
                active_mons = private_reserve_x[active]

                moves = active_mons[..., -8:]
                moves = moves.view(*moves.shape[:-1], 4, 2)
                move_tokens = moves[..., 0]
                move_names = self.move_embedding.get_name(move_tokens)
                moves_emb = self.move_embedding(move_tokens)
                moves_basepower = moves_emb[..., 3].squeeze()
                moves_basepower = torch.masked_fill(
                    moves_basepower, ~state["move_mask"].squeeze(), -1
                )

                index = torch.argmax(moves_basepower, -1).item()
                data = choices["moves"]
                func, args, kwargs = data[index]
            else:
                random_key = random.choice(
                    [key for key, value in choices.items() if value]
                )
                _, (func, args, kwargs) = random.choice(
                    list(choices[random_key].items())
                )
        except:
            random_key = random.choice([key for key, value in choices.items() if value])
            _, (func, args, kwargs) = random.choice(list(choices[random_key].items()))
        finally:
            return func, args, kwargs



import torch
import asyncio

from typing import Any, Sequence, Mapping, Type

from meloetta.player import Player
from meloetta.room import BattleRoom
from meloetta.actors.base import Actor
from meloetta.workers.barrier import Barrier

from meloetta.room import BattleRoom
from meloetta.utils import expand_bt

DRAW_BY_TURNS = 300


def waiting_for_opp(room: BattleRoom):
    controls = room.get_js_attr("controls.controls")
    return (
        "Waiting for opponent" in controls or " will switch in, replacing" in controls
    )


class SelfPlayWorker:
    def __init__(
        self,
        worker_index: int,
        num_players: int,
        battle_format: str,
        team: str,
        actor_fn: Type[Actor],
        actor_args: Sequence[Any] = None,
        actor_kwargs: Mapping[str, Any] = None,
    ):
        self.battle_format = battle_format
        self.team = team

        self.worker_index = worker_index
        self.num_players = num_players

        self.actor_fn = actor_fn
        self.actor_args = () if actor_args is None else actor_args
        self.actor_kwargs = {} if actor_kwargs is None else actor_kwargs

    def __repr__(self) -> str:
        return f"Worker{self.worker_index}"

    def run(self) -> Any:
        """
        Start selfplay between two asynchronous actors-
        """

        async def selfplay():
            barrier = Barrier(self.num_players)
            return await asyncio.gather(
                *[
                    self.actor(self.worker_index * self.num_players + i, barrier)
                    for i in range(self.num_players)
                ]
            )

        results = asyncio.run(selfplay())
        return results

    async def actor(self, player_index: int, barrier: Barrier) -> Any:
        username = f"player{player_index}"

        player = await Player.create(username, None, "localhost:8000")
        await player.client.login()
        await barrier.wait()

        while True:  # 10 battles each player-player pair
            if player_index % 2 == 0:
                await player.client.challenge_user(
                    f"player{player_index + 1}", self.battle_format, self.team
                )
            else:
                await player.client.accept_challenge(self.battle_format, self.team)

            turn = 0
            turns_since_last_move = expand_bt(torch.tensor(0, dtype=torch.long))
            actor = self.actor_fn(
                *self.actor_args,
                **self.actor_kwargs,
                pid=(0 if player_index % 2 == 0 else 1),
            )

            while True:
                message = await player.client.receive_message()
                action_required = await player.recieve(message)
                if "is offering a tie." in message:
                    await player.client.websocket.send(
                        player.room.battle_tag + "|" + "/offertie"
                    )
                if "|error" in message:
                    # edge case for handling when the pokemon is trapped
                    if "Can't switch: The active Pokémon is trapped" in message:
                        message = await player.client.receive_message()
                        action_required = await player.recieve(message)
                        action_required = True

                    # for some reason, disabled max moves are being selected
                    elif "Can't move" in message:
                        print(message)

                    else:
                        print(message)

                if action_required:
                    # inputs to neural net
                    battle = player.room.get_battle()
                    turn = battle["turn"]
                    vstate = actor.get_vectorized_state(player.room, battle)

                ended = player.room.get_js_attr("battle?.ended")
                while (
                    action_required and not waiting_for_opp(player.room) and not ended
                ):
                    choices = player.get_choices(turns_since_last_move)
                    state = {
                        **vstate,
                        **choices.action_masks,
                        **choices.prev_choices,
                        "targeting": choices.targeting,
                    }

                    func, args, kwargs = actor(state, player.room, choices.choices)
                    func(*args, **kwargs)

                outgoing_message = player.room.get_js_attr("outgoing_message")
                if outgoing_message:
                    player.room.pop_outgoing()
                    if "move" in outgoing_message:
                        turns_since_last_move = turns_since_last_move * 0
                    else:
                        turns_since_last_move = turns_since_last_move + 1
                    await player.client.websocket.send(
                        player.room.battle_tag + "|" + outgoing_message
                    )

                if ended:
                    break

                if turn > DRAW_BY_TURNS:
                    print(f"{username}: draw by turn > {DRAW_BY_TURNS}!")

                    await player.client.websocket.send(
                        player.room.battle_tag + "|" + "/offertie"
                    )
                    while True:
                        message = await player.client.receive_message()
                        action_required = await player.recieve(message)
                        if "is offering a tie." in message:
                            await player.client.websocket.send(
                                player.room.battle_tag + "|" + "/offertie"
                            )
                        ended = player.room.get_js_attr("battle?.ended")
                        if ended:
                            break
                    break

            await player.client.leave_battle(player.room.battle_tag)
            actor.post_match(player.room)
            player.reset()



import os
import re
import json

from typing import Union, Dict, Any, List
from py_mini_racer import MiniRacer
from py_mini_racer.py_mini_racer import is_unicode
from meloetta.data import ROOT_DIR

SRC = [
    f"{ROOT_DIR}/js/predef.js",
    f"{ROOT_DIR}/js/polyfill.js",
    f"{ROOT_DIR}/js/cycle.js",
    f"{ROOT_DIR}/js/underscore.js",
    f"{ROOT_DIR}/js/client/aliases.js",
    f"{ROOT_DIR}/js/client/abilities.js",
    f"{ROOT_DIR}/js/client/items.js",
    f"{ROOT_DIR}/js/client/moves.js",
    f"{ROOT_DIR}/js/client/pokedex.js",
    f"{ROOT_DIR}/js/client/typechart.js",
    f"{ROOT_DIR}/js/client/formats-data.js",
    f"{ROOT_DIR}/js/client/teambuilder-tables.js",
    f"{ROOT_DIR}/js/client/battle-scene-stub.js",
    f"{ROOT_DIR}/js/client/battle-choices.js",
    f"{ROOT_DIR}/js/client/battle-dex.js",
    f"{ROOT_DIR}/js/client/battle-dex-data.js",
    f"{ROOT_DIR}/js/client/battle-text-parser.js",
    f"{ROOT_DIR}/js/client/battle-tooltips.js",
    f"{ROOT_DIR}/js/client/battle-log.js",
    f"{ROOT_DIR}/js/client/battle.js",
    f"{ROOT_DIR}/js/client.js",
    f"{ROOT_DIR}/js/engine.js",
]


def deserialize(state: Dict[str, Union[str, List[str], Dict[str, Any]]]):
    def rez(value):
        nonlocal state
        if isinstance(value, list):
            for i, element in enumerate(value):
                if (
                    isinstance(element, dict) or isinstance(element, list)
                ) and element is not None:
                    try:
                        path = element.get("$ref")
                    except:
                        path = None
                    if isinstance(path, str):
                        path = path.replace("$", "state")
                        value[i] = eval(path)
                    else:
                        rez(element)
        elif isinstance(value, dict):
            for name in value.keys():
                item = value[name]
                if (
                    isinstance(item, dict) or isinstance(item, list)
                ) and item is not None:
                    try:
                        path = item.get("$ref")
                    except:
                        path = None
                    if isinstance(path, str):
                        path = path.replace("$", "state")
                        value[name] = eval(path)
                    else:
                        rez(item)

    rez(state)
    return state


class BattleRoom:
    title: str

    def __init__(self):
        self._ctx = MiniRacer()
        self._battle_tag = None
        self.myPokemon = None
        self.request = None
        self.ended = False
        self._load_js()

    def _load_js(self):
        for file in SRC:
            with open(file, "r", encoding="utf-8") as f:
                file_src = f.read()
            try:
                modules = re.findall(r"(module\.exports \= .*)", file_src)
                for module in modules:
                    file_src = file_src.replace(module, "")
                self._ctx.eval(file_src)
            except Exception as e:
                print(e)
                print(file)
                exit()
        self._ctx.eval("engine.start()")

    def _call(self, cmd, *args):
        js = cmd + "({})".format(json.dumps(args)[1:-1])
        return self._execute(js)

    def _execute(self, expr, timeout=None, max_memory=None):
        wrapped_expr = "JSON.stringify((function(){return (%s)})())" % expr
        ret = self._ctx.eval(wrapped_expr, timeout=timeout, max_memory=max_memory)
        if not is_unicode(ret):
            return None
        return self._ctx.json_impl.loads(ret)

    def recieve(self, data: str = ""):
        return self._execute("engine.receive({})".format(json.dumps(data)))

    def get_js_attr(self, attr: str):
        return self._execute(f"engine.client.{attr}")

    def get_battle(self, raw: bool = False):
        battle = self._execute("engine.serializeBattle()")
        if not raw:
            battle = deserialize(battle)
        return battle

    def get_species(self, species):
        return self._call("engine.getSpecies", species)

    def get_move(self, move):
        return self._call("engine.getMove", move)

    def get_item(self, item):
        return self._call("engine.getItem", item)

    def get_ability(self, ability):
        return self._call("engine.getAbility", ability)

    def set_gen(self, gen):
        return self._call("engine.setGen", gen)

    def get_pid(self):
        return self._execute("engine.getPid()")

    def get_reward(self):
        return self._execute("engine.getReward()")

    def reset(self):
        self._ctx.eval("engine.reset()")
        self._battle_tag = None

    # choice start

    def choose_move_target(self, posString):
        return self._execute(
            "engine.chooseMoveTarget({})".format(json.dumps(posString))
        )

    def choose_move(
        self,
        pos: str,
        target: str,
        isMega: bool = False,
        isZMove: bool = False,
        isUltraBurst: bool = False,
        isDynamax: bool = False,
        isTerastal: bool = False,
    ):
        args = [pos, target, isMega, isZMove, isUltraBurst, isDynamax, isTerastal]
        args = json.dumps(args)[1:-1]
        cmd = "engine.chooseMove({})".format(args)
        return self._execute(cmd)

    def choose_shift(self):
        return self._execute("engine.chooseShift()")

    def choose_switch(self, pos: str):
        return self._execute("engine.chooseSwitch({})".format(json.dumps(pos)))

    def choose_switch_target(self, pos: str):
        return self._execute("engine.chooseSwitchTarget({})".format(json.dumps(pos)))

    def choose_team_preview(self, pos: str):
        return self._execute("engine.chooseTeamPreview({})".format(json.dumps(pos)))

    def pop_outgoing(self):
        return self._execute("engine.popOutgoing()")

    # choice end

    def get_state(self, raw: bool = True) -> Dict[str, Union[str, Dict[str, Any]]]:
        state = self._execute("engine.serialize()")
        if not raw:
            state = deserialize(state)
        return state

    @property
    def battle_tag(self):
        return self._battle_tag


def main():
    room = BattleRoom()


if __name__ == "__main__":
    main()



from meloetta.frameworks.porygon.model.encoder import Encoder
from meloetta.frameworks.porygon.model.policy import PolicyHeads
from meloetta.frameworks.porygon.model.core import Core
from meloetta.frameworks.porygon.model.value import ValueHead
from meloetta.frameworks.porygon.model.heads.move_head import MoveHead
from meloetta.frameworks.porygon.model.heads.max_move_head import MaxMoveHead
from meloetta.frameworks.porygon.model.heads.switch_head import SwitchHead
from meloetta.frameworks.porygon.model.heads.flag_head import FlagsHead
from meloetta.frameworks.porygon.model.interfaces import *
from meloetta.frameworks.porygon.model.utils import *
from meloetta.frameworks.porygon.model.main import PorygonModel
from meloetta.frameworks.porygon.model.config import *



import torch
import torch.nn as nn
import numpy as np

from typing import Dict, Tuple, Any, List


Action = Any


class Node(object):
    def __init__(self, prior: float):
        self.visit_count = 0
        self.to_play = -1
        self.prior = prior
        self.value_sum = 0
        self.children = {}
        self.hidden_state = None
        self.reward = 0

    def expanded(self) -> bool:
        return len(self.children) > 0

    def value(self) -> float:
        if self.visit_count == 0:
            return 0
        return self.value_sum / self.visit_count


def ucb_scores(
    parent: Node, children: List[Node], pb_c_base: int, pb_c_init: int
) -> Tuple[Action, Node]:
    child_priors = torch.tensor([child.prior for child in children])
    child_visit_counts = torch.tensor([child.visit_count for child in children])

    pb_c = np.log((parent.visit_count + pb_c_base + 1) / pb_c_base) + pb_c_init
    pb_c *= np.sqrt(parent.visit_count) / (child_visit_counts + 1)

    prior_scores = pb_c * child_priors

    child_value_sums = torch.tensor([child.value_sum for child in children])
    value_scores = child_value_sums / child_visit_counts.clamp(min=1)
    value_scores *= child_visit_counts > 0

    return prior_scores + value_scores


class Tree:
    def __init__(self, model: nn.Module, num_simulations: int = 50):
        self.model = model
        self.root = Node(0)
        self.num_simulations = num_simulations

    def select_child(self, node: Node) -> Tuple:
        actions = list(node.children.keys())
        children = list(node.children.values())

        scores = ucb_scores(node, children)
        _, action, child = torch.argmax(scores)

        return action, child

    def run(self, initial_observation: Dict[str, torch.Tensor]):

        for _ in range(self.num_simulations):
            history = []
            node = self.root
            search_path = [node]

            while node.expanded():
                action, node = self.select_child(node)



from typing import NamedTuple


from meloetta.data import VOLATILES, SIDE_CONDITIONS, WEATHERS, PSEUDOWEATHERS


class _Config(NamedTuple):
    @classmethod
    def from_dict(self, d: dict):
        return self(**d)


class SideEncoderConfig(_Config):
    entity_embedding_dim: int = 256

    transformer_input_size: int = entity_embedding_dim
    transformer_num_heads: int = 2
    transformer_num_layers: int = 3
    resblocks_num_before: int = 1
    resblocks_num_after: int = 1
    resblocks_hidden_size: int = 256

    output_dim: int = 256


class ScalarEncoderConfig(_Config):
    embedding_dim: int = 64


class WeatherEncoderConfig(_Config):
    embedding_dim: int = 64


class EncoderConfig(_Config):
    side_encoder_config: SideEncoderConfig = SideEncoderConfig()
    scalar_encoder_config: ScalarEncoderConfig = ScalarEncoderConfig()
    weather_encoder_config: WeatherEncoderConfig = WeatherEncoderConfig()
    output_dim: int = (
        3 * side_encoder_config.output_dim
        # + 2 * (8 * 6 + len(VOLATILES) + (2 + 2 + 3 + (len(SIDE_CONDITIONS) - 4)))
        + 444
        # + 3
        # + (
        #     (len(WEATHERS) + 1)
        #     + 10
        #     + 7
        #     + 8 * len(PSEUDOWEATHERS)
        #     + 10 * len(PSEUDOWEATHERS)
        # )
        + 85
        + 242
    )


class CoreConfig(_Config):
    raw_embedding_dim: int = EncoderConfig.output_dim
    hidden_dim: int = 512
    num_layers: int = 1


class ValueHeadConfig(_Config):
    state_embedding_dim: int = CoreConfig.hidden_dim
    hidden_dim: int = 256
    num_resblocks: int = 1


class ActionTypeHeadConfig(_Config):
    state_embedding_dim: int = CoreConfig.hidden_dim
    num_action_types: int = 3
    context_dim: int = ScalarEncoderConfig.embedding_dim
    residual_dim: int = 256
    action_map_dim: int = 256
    autoregressive_embedding_dim: int = 256


class FlagHeadConfig(_Config):
    autoregressive_embedding_dim: int = CoreConfig.hidden_dim
    hidden_dim: int = 256


class MaxMoveHeadConfig(_Config):
    entity_embedding_dim: int = SideEncoderConfig.entity_embedding_dim
    key_dim: int = 64

    autoregressive_embedding_dim: int = CoreConfig.hidden_dim
    query_hidden_dim: int = 256


class MoveHeadConfig(_Config):
    entity_embedding_dim: int = SideEncoderConfig.entity_embedding_dim
    key_dim: int = 64

    autoregressive_embedding_dim: int = CoreConfig.hidden_dim
    query_hidden_dim: int = 256


class SwitchHeadConfig(_Config):
    entity_embedding_dim: int = SideEncoderConfig.entity_embedding_dim
    key_dim: int = 64

    autoregressive_embedding_dim: int = CoreConfig.hidden_dim
    query_hidden_dim: int = 256


class TargetHeadConfig(_Config):
    pass


class PolicyHeadsConfig(_Config):
    action_type_head_config: ActionTypeHeadConfig = ActionTypeHeadConfig()
    flag_head_config: FlagHeadConfig = FlagHeadConfig()
    max_move_head_config: MaxMoveHeadConfig = MaxMoveHeadConfig()
    move_head_config: MoveHeadConfig = MoveHeadConfig()
    switch_head_config: SwitchHeadConfig = SwitchHeadConfig()
    target_head_config: TargetHeadConfig = TargetHeadConfig()

    entity_embedding_dim: int = SideEncoderConfig.entity_embedding_dim
    key_dim: int = MoveHeadConfig.key_dim

    autoregressive_embedding_dim: int = CoreConfig.hidden_dim
    query_hidden_dim: int = 32


def _scale(config: _Config, factor: float):
    datum = {}
    for key, value in config._asdict().items():
        if isinstance(value, tuple) and hasattr(value, "_asdict"):
            datum[key] = _scale(value, factor)
        elif isinstance(value, int) or isinstance(value, float):
            datum[key] = max(int(value * factor), 1)

    return config.from_dict(datum)


class PorygonModelConfig(_Config):
    encoder_config: EncoderConfig = EncoderConfig()
    policy_heads_config: PolicyHeadsConfig = PolicyHeadsConfig()
    resnet_core_config: CoreConfig = CoreConfig()
    value_head_config: ValueHeadConfig = ValueHeadConfig()

    @classmethod
    def from_scale(self, scale: float):
        return _scale(self(), scale)


def main():
    config = PorygonModelConfig.from_scale(0.2)
    print(config)


if __name__ == "__main__":
    main()
